/****************************************************************************
 MADE BY BEN. BEN IS COOL. Feb 6, 2017
 Copyright Noodlecake Studios Inc. All rights reserved.
 ****************************************************************************/

#include "NoodleBinkController.h"
#include "bink.h"
#include "binktextures.h"


static NoodleBinkController *staticController = NULL;

static HBINK Bink;

static BINKTEXTURES * textures;
static BINKSHADERS * shaders;

static double x0;
static double y0;
static double x1;
static double y1;

static bool usingObb;

static bool initialized;

static char* currentVideo;

static std::string lastFilePath;
static long lastFileOffset;
std::string obbPath;
NoodleBinkController* NoodleBinkController::sharedController()
{
    if (staticController == NULL)
    {
        staticController = new NoodleBinkController();
    }
    return staticController;
}

int NoodleBinkController::initBink(AAssetManager *mgr, bool useObb, std::string _obbPath)
{
    Bink = 0;
    textures = 0;
    shaders = 0;
    lastFilePath = "";
    usingObb = useObb;
    obbPath = _obbPath;
    currentVideo = NULL;

    // should be something like "/mnt/sdcard/Android/obb/air.com.noodlecake.luminocity/main.1000001.air.com.noodlecake.luminocity.obb"
    // __android_log_print(ANDROID_LOG_ERROR, NOODLEBINK_LOG_TAG, "[NoodleBinkController] obbPath: %s", obbPath.c_str());

    // Spin off another thread to decode our Binks. 
    BinkStartAsyncThread( 0, 0 );

    // Set the asset manager
    BinkSetAssetManager(mgr);

    // Create the shaders
    shaders = Create_Bink_shaders(0);
    if (!shaders) {
        __android_log_print(ANDROID_LOG_ERROR, NOODLEBINK_LOG_TAG, "Failed to initialize opengl shaders.");
        return 1;
    }

    initialized = true;

    return 0;
}

void NoodleBinkController::reinitBink(AAssetManager *mgr)
{
    if (!initialized) {
        __android_log_print(ANDROID_LOG_ERROR, NOODLEBINK_LOG_TAG, "Failed to reinitialize Bink: Bink not initialized for the first time");
        return;
    }

    int lastFrame = 0;
    
    // close the bink
    if (Bink) {
        lastFrame = (int)(Bink->LastFrameNum);
        BinkDoFrameAsyncWait( Bink, -1 );
        BinkClose(Bink);
        Bink = 0;
    }

    // free the texture memory
    if (textures) {
        Free_Bink_textures(textures);
        textures = 0;
    }

    if (currentVideo != NULL) {
        free(currentVideo);
        currentVideo = NULL;
    }

    if (shaders) {
        Free_Bink_shaders(shaders);
        shaders = 0;
    }
    
    std::string lastPathRetain = lastFilePath;

    // reinitialize
    initBink(mgr, usingObb, obbPath);

    //reload video that was in memory and go to that frame
    if (lastPathRetain.compare("") != 0) {
        //there was already a video loaded previously, load it
        loadVideo(lastPathRetain, lastFileOffset);
        gotoFrame(lastFrame);
    }
}

int NoodleBinkController::loadVideo(std::string path, long offset) {
    if (!initialized) {
        __android_log_print(ANDROID_LOG_ERROR, NOODLEBINK_LOG_TAG, "Failed to open Bink file: Bink not initialized");
        return 1;
    }

    lastFilePath = path;
    lastFileOffset = offset;

    BinkSetSoundTrack(0,0);
    
    if(Bink)
    {
        BinkClose(Bink);
    }

    if (textures) {
        Free_Bink_textures(textures);
    }

    if (currentVideo != NULL) {
        free(currentVideo);
        currentVideo = NULL;
    }

    if (usingObb) {
        BinkSetFileOffset((U64)offset);
        Bink = BinkOpen(obbPath.c_str(), BINKFILEOFFSET);
    } else {
        Bink = BinkOpen(path.c_str(), BINKNOSKIP);
    }
    if (!Bink)
    {
        __android_log_print(ANDROID_LOG_ERROR, NOODLEBINK_LOG_TAG, "Failed to open Bink file: %s", BinkGetError());
        return 1;
    }

    // Create the textures
    textures = Create_Bink_textures(shaders, Bink, 0);
    if(!textures) {
        __android_log_print(ANDROID_LOG_ERROR, NOODLEBINK_LOG_TAG, "Failed to create bink textures");
        return 1;
    } else {
        Set_Bink_draw_position(textures, x0, y0, x1, y1);
    }
    
    // Start decoding the first frame.
    // (The logic in update() expects the first frame to be decoding.)
    Start_Bink_texture_update(textures);
    BinkDoFrameAsync(Bink, 0, 0);

    return 0;
}

int NoodleBinkController::getTotalFrames() {
    if (!initialized || Bink == 0) {
         __android_log_print(ANDROID_LOG_ERROR, NOODLEBINK_LOG_TAG, "Failed to get total frames: Bink not initialized");
        return 0;
    }

    BINKSUMMARY summary;
    BinkGetSummary(Bink, &summary);
    return summary.TotalFrames;
}

bool NoodleBinkController::nextFrame() {
    if (!initialized || Bink == 0) {
         __android_log_print(ANDROID_LOG_ERROR, NOODLEBINK_LOG_TAG, "Failed to get next frame: Bink not initialized");
        return false;
    }
    
    if (BinkDoFrameAsyncWait(Bink, -1))
    {
        BinkNextFrame(Bink);

        Finish_Bink_texture_update(textures);

        Start_Bink_texture_update(textures);
        BinkDoFrameAsync(Bink, 0, 0);
        
        return true;
    } else {
        return false;
    }
}

bool NoodleBinkController::gotoFrame(int frame) {
    if (!initialized || Bink == 0)
    {
         __android_log_print(ANDROID_LOG_ERROR, NOODLEBINK_LOG_TAG, "Failed to go to frame: Bink not initialized");
        return false;
    }
        
    BinkDoFrameAsyncWait( Bink, -1 );
        

    BinkGoto(Bink, frame,0);
        
    Finish_Bink_texture_update(textures);
    Start_Bink_texture_update(textures);
    BinkDoFrameAsync(Bink, 0, 0);
        
    return true;
}

void NoodleBinkController::setdraw(double nx0, double ny0, double nx1, double ny1) {
    x0 = nx0;
    y0 = ny0;
    x1 = nx1;
    y1 = ny1;

    if (textures) {
        Set_Bink_draw_position(textures, x0, y0, x1, y1);
    }
}

void NoodleBinkController::draw() {
    if (!initialized || textures == 0) {
        __android_log_print(ANDROID_LOG_ERROR, NOODLEBINK_LOG_TAG, "Failed to draw: Bink not initialized");
        return;
    }

    Draw_Bink_textures(textures, 0, 0);
}
