/****************************************************************************
//  Created by Ben Schmidt on Jan 24, 2017
//  Copyright (c) 2017 Noodlecake Studios Inc. All rights reserved.
 ****************************************************************************/

package com.noodlecake.BinkANE;

import com.adobe.fre.FREContext;
import com.noodlecake.BinkANE.Extension;
import android.util.Log;

import android.app.ActivityManager;
import android.app.ActivityManager.MemoryInfo;
import android.graphics.Point;
import android.graphics.Color;
import android.graphics.PixelFormat;
import com.noodlecake.NoodleX.NoodleXOBB;
import android.view.Display;
import android.view.Gravity;
import android.view.KeyEvent;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.view.TextureView;
import android.view.View;
import android.view.ViewGroup;
import android.view.Window;
import android.view.animation.Animation;
import android.view.animation.ScaleAnimation;
import android.view.animation.TranslateAnimation;
import android.view.animation.AccelerateDecelerateInterpolator;
import android.view.animation.AccelerateInterpolator;
import android.view.animation.AlphaAnimation;
import android.view.animation.AnimationSet;
import android.view.animation.LinearInterpolator;
import android.view.animation.RotateAnimation;

import android.media.MediaMetadataRetriever;
import android.media.MediaPlayer;

import android.net.Uri;

import android.content.pm.ActivityInfo;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.pm.PackageManager.NameNotFoundException;

import android.support.annotation.MainThread;
import android.support.v4.view.ViewGroupCompat;

import android.widget.FrameLayout;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.MediaController;
import android.widget.RelativeLayout;
import android.widget.VideoView;
import android.widget.ImageView.ScaleType;
import android.widget.ImageButton;

import android.R.transition;

import android.animation.Animator;
import android.animation.AnimatorListenerAdapter;

import android.app.Activity;

import android.content.Context;
import android.content.res.AssetManager;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;

import android.graphics.Bitmap;
import android.graphics.BitmapShader;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.ColorMatrix;
import android.graphics.ColorMatrixColorFilter;
import android.graphics.Paint;
import android.graphics.PorterDuffColorFilter;
import android.graphics.PorterDuff.Mode;
import android.graphics.Shader.TileMode;
import android.graphics.PorterDuffXfermode;
import android.graphics.Matrix.ScaleToFit;
import android.graphics.drawable.Drawable;

import java.lang.String;

import java.security.KeyStore.TrustedCertificateEntry;

import java.util.Map;
import java.util.concurrent.Exchanger;
import java.util.concurrent.Callable;
import java.util.HashMap;

import java.io.InputStream;
import java.io.File;
import java.io.FileOutputStream;

import com.noodlecake.NoodleX.NoodleXBridge;

/* obb stuff */
import com.android.vending.expansion.zipfile.APKExpansionSupport;
import com.android.vending.expansion.zipfile.ZipResourceFile;
import com.android.vending.expansion.zipfile.ZipResourceFile.ZipEntryRO;
/* end obb stuff */


public class BinkController
{
    private static Activity sharedActivity;

    private static FREContext context;
    private static int loadedChapters;
    
    private static int currentChapterPosition;
    
    private static boolean showLoadingCircle;
    
//     UIImageView* screenShotHolder;
//     UIView* whiteFlash;
//     UIImage* lastSavedImage;
//     UIView* sharePosition;
    
    private static SurfaceView mainView;
    private static FrameLayout creditsLayout;
    private static ImageButton videoHomeButton;
    private static ImageButton videoMenuButton;
    private static VideoView videoView;
    private static ImageView logoImage;
    private static ImageView fadeView;
    private static ImageView loadingView;
    private static ImageView loadingImage;
    private static ImageView obbLoadingImage;

    private static MediaPlayer firstAudioPlayer;
    
    private static boolean didAuthGPGS;

    private static boolean loadingObbView = true;

    private static HashMap<String, Bitmap> imagesLoaded;

    private static SharedPreferences preferences;
    private static SharedPreferences.Editor prefEditor;
    private static String chapterKey = "chapter_";              // A profiles chapter key, concatinated with a profile number
    private static String updatedKey = "updated_";              // The time stamp when a profile was updated, concatinated with a profile number
    private static String tipsKey = "tips_";                    // Whether or not a profile should show tips, concatinated with a profile number
    private static String dataKey = "data_";                    // The data associated with chapter progressions for a profile, concatinated with a profile number
    private static String selectedProfile = "selectedProfile";  // The integer for the selected profile.
    private static String hasRated = "didrate1";                // Exists when the show rating has occured. Doesn't exist when hasn't been prompted to rate.

    private static String mCloudSaveString = "";                                // Game-side cloud save string. 
    private static HashMap<String,String> mCloudSaveMap = new HashMap<>();;     // The converted cloud save string -> hashmap. NOTE: GAME SPECIFIC.

    private static String nextLoadDir;
    private static String nextLoadVideo;
    private static int loadedStage;

    private static boolean gpgsSignInComplete = false;                      // Have we tried to sign into google play yet?
    private static boolean gpgsCloudSaveComplete = false;                   // Have we tried to get a cloud save yet?
    private static boolean obbComplete = false;                             // Have we checked on our assets and got an obb if necessary?

    private static final String obbTestFile = "scenes/Scene_Boat.swf";      // Random file that would generally be packaged into the obb
    private static boolean usingObb = false;                                // Is the game loading from obb or directly from the apk?

    private static int memAvailable;

    /*** INIT ***/
    public static boolean icInitialise(FREContext airContext) {
        context = airContext;
        sharedActivity = context.getActivity();

        // Set up preferences stuff
        if (preferences == null) preferences = sharedActivity.getPreferences(Context.MODE_PRIVATE);
        if (prefEditor == null) prefEditor = preferences.edit();

        // Set Noodlenews creative callback
        NoodleXBridge.setCreativesCallable(new Callable<Void>()
            {
                public Void call()
                {   
                    noodlenewsMuteAudio();
                    return null;
                }
            });

        // Set up main view
        mainView = (SurfaceView)(sharedActivity.getCurrentFocus());

        loadedStage = 0;
        nextLoadVideo = "";
        nextLoadDir = "";

        showLoadingCircle = true;

        currentChapterPosition = 1;
        
        imagesLoaded = new HashMap<String, Bitmap>();
        
        didAuthGPGS = false;

        NoodleXBridge.setLocalization(NoodleXBridge.getLocalizationCode());
        
        System.loadLibrary("NoodleBink");

        ActivityManager activityManager = (ActivityManager) sharedActivity.getSystemService(Context.ACTIVITY_SERVICE);
        memAvailable = activityManager.getLargeMemoryClass();
        Log.d(Extension.EXT_LOG_TAG, "Device memory available: " + memAvailable);

        return true;
    }

    public static void checkStartupData() {
        //request cloud save
        //sign into google play, we'll try to get cloud save data once it's signed in
        NoodleXBridge.setAuthenticateCompleteCallback(new Callable<Void>()
        {
            public Void call()
            {   
                gpgsSignInComplete = true;
                startupStepComplete();
                return null;
            }
        });
        authenticateLocalUser();

        //see if the assets are already packed into the apk, or if we're using an obb
        try { 
            AssetManager assetManager = sharedActivity.getAssets();
            InputStream stream = assetManager.open(obbTestFile);
            
            //if opening didn't throw an exception, then we have the file in the apk
            stream.close();
            usingObb = false;

        } catch (Exception e) {
            //file couldn't be found in the apk, we need to use an obb
            usingObb = true;
        }

        //if we're using an obb, check if we have that obb and download it if we don't
        if (usingObb) {

            //set obb download callbacks
            NoodleXBridge.setObbDownloadStartedCallback(new Callable<Void>()
            {
                public Void call()
                {   
                    startLoadingObbView();
                    return null;
                }
            });

            NoodleXBridge.setObbDownloadCompleteCallback(new Callable<Void>()
            {
                public Void call()
                {   
                    if (NoodleXBridge.getObbDownloadState() == NoodleXBridge.NOODLEX_SUCCESS) {
                        //I guess we don't need to do anything here anymore
                    } else {
                        //if the obb failed to download, the user will be prompted by NoodleX to restart the app
                        //the obbPreparedCallback will handle closing the app
                    }
                    return null;
                }
            });

            NoodleXBridge.setObbPreparedCallback(new Callable<Void>()
            {
                public Void call()
                {   
                    if (NoodleXBridge.obbReady()) {
                        obbComplete = true;
                        startupStepComplete();
                    } else {
                        //only way we should get here is if the user cancelled the permission or something went wrong
                        //and if something went wrong NoodleXOBB should have alerted the user.
                        Log.e(Extension.EXT_LOG_TAG, "Obb error or permissions not granted. Exiting game.");
                        context.dispatchStatusEventAsync("EXITGAME", "TRUE");
                    }

                    return null;
                }
            });

            //download the obb
            NoodleXBridge.prepareObb();
        } else {
            //files are in the apk - good to go!
            obbComplete = true;
            startupStepComplete();
        }
    }

    private static void startupStepComplete() {
        Log.d(Extension.EXT_LOG_TAG, "Startup step complete! gpgsSignInComplete: " + gpgsSignInComplete + "   gpgsCloudSaveComplete: " + gpgsCloudSaveComplete + "   obbComplete: " + obbComplete);
        if (gpgsSignInComplete && !gpgsCloudSaveComplete) {
            //if we correctly signed in, check for cloud sync
            if (NoodleXBridge.isAuthenticatedGPGS()) {
                NoodleXBridge.setCloudLoadCallback(new Callable<Void>()
                {
                    public Void call()
                    {   
                        gpgsCloudSaveComplete = true;
                        try {
                            updateLocalCloudSaveData();
                        } catch (Exception e) {
                            Log.e(Extension.EXT_LOG_TAG, "Error updating cloud save: " + e.toString());
                        }
                        startupStepComplete();
                        return null;
                    }
                });
                NoodleXBridge.loadFromCloud(0);
            } else {
                gpgsCloudSaveComplete = true;
            }
        } 
        
        if (gpgsSignInComplete && gpgsCloudSaveComplete && obbComplete) {

            if (NoodleXBridge.getLaunchCount() == 1) {
                NoodleXBridge.showToast(NoodleXBridge.getLocalizedMessage("NC_PreparingFirstLaunch"));
            }

            //wanna make sure there's enough room for the outro before people get there...
            unpackAssetToCache("ending/Outro.mp4", usingObb);
            
            Log.d(Extension.EXT_LOG_TAG, "All startup steps complete.");
            context.dispatchStatusEventAsync("STARTUPDATAAVAILABLE", "TRUE");
        }
    }

    public static boolean usingObb() {
        return usingObb;
    }


    /*** ASSET LOADING ***/
    public static String getAssetUrl(String name, String ext) {
        //not implemented - not currently used on Android
        //may want to eventually use this to decide between assets based on device performance or screen size
        return null;
    }

    public static String getAssetUrlInDir(String dir, String name, String ext) {
        //not implemented - not currently used on Android
        //may want to eventually use this to decide between assets based on device performance or screen size
        return null;
    }

    public static void loadImageFor(final String name) {
        Thread t = new Thread(new Runnable() {
             public void run() {
                String rgbName = name + "_RGB/image.jpg";
                String aName = name + "_A/image.jpg";

                AssetManager mgr = sharedActivity.getAssets();

                InputStream inStream;
                Bitmap rgbImg = null;
                Bitmap aImg = null;

                try {
                    if (usingObb) {
                        inStream = NoodleXBridge.getObbInputStreamForFile(rgbName);
                    } else {
                        inStream = mgr.open(rgbName);
                    }
                    rgbImg = BitmapFactory.decodeStream(inStream);
                    
                    if (usingObb) {
                        inStream = NoodleXBridge.getObbInputStreamForFile(aName);
                    } else {
                        inStream = mgr.open(aName);
                    }
                    aImg = BitmapFactory.decodeStream(inStream);

                } catch (Exception e) {
                    Log.e(Extension.EXT_LOG_TAG, "Failed to load image: " + name);
                    return;
                }

                float[] matrix = new float[] {
                    0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0,
                    -1, 0, 0, 0, 255};

                Paint alphaPaint = new Paint();
                alphaPaint.setColorFilter(new ColorMatrixColorFilter(new ColorMatrix(matrix)));

                //Useful link - http://ssp.impulsetrain.com/porterduff.html
                alphaPaint.setXfermode(new PorterDuffXfermode(Mode.DST_IN));

                //FYI, if you ever want to use an operation that directly sets a color on the bitmap,
                //the canvas utilities take in ARGB but AIR expects ABGR, so manually create a color
                //using Color.argb(A, B, G, R)

                Bitmap result = rgbImg.copy(Bitmap.Config.ARGB_8888, true);
                Canvas c = new Canvas(result);
                c.setDensity(Bitmap.DENSITY_NONE);
                c.drawBitmap(aImg, 0, 0, alphaPaint);
                
                imagesLoaded.put(name, result);

                context.dispatchStatusEventAsync("LOADED_IMAGE", name);
            }
        });  
        t.start();
    }

    public static Bitmap getLoadedImage(String name) {
        return imagesLoaded.get(name);
    }  

    public static void deleteLoadedImage(String name) {
        imagesLoaded.remove(name);
    }

    public static byte[] getBytesFor(String name) {
        
        InputStream stream;

        if (!usingObb) {
            try { 
                AssetManager assetManager = sharedActivity.getAssets();
                stream = assetManager.open(name);
                byte[] data = new byte[stream.available()];
                stream.read(data);
                stream.close();
                return data;
            } catch (Exception e) {
                Log.e(Extension.EXT_LOG_TAG, "Failed to get bytes for file " + name);
                return null;
            }
        } else {
            return NoodleXBridge.getObbBytesForFile(name);
        }
    }

    //returns a path to the cached file, or null if something went wrong
    public static String unpackAssetToCache(String assetPath, boolean inObb) {
        AssetManager mgr = sharedActivity.getAssets();
        
        //create new file
        File unpacked = new File(sharedActivity.getCacheDir() + "/" + assetPath);

        //make sure releavant parent folders exist
        File parent = unpacked.getParentFile();
        if(!parent.exists() && !parent.mkdirs()){
            Log.e(Extension.EXT_LOG_TAG, "Failed to create cache file directories for file " + assetPath);
            return null;
        }

        InputStream inStream = null;
        FileOutputStream outStream = null;
        try {
            if (!inObb) {
                inStream = sharedActivity.getAssets().open(assetPath);
            } else {
                inStream = NoodleXBridge.getObbInputStreamForFile(assetPath);
            }
            long size = inStream.available();

            long existingFileSize = 0;
            if (unpacked.exists()) {
                Log.d(Extension.EXT_LOG_TAG, "File already in the cache!");
                existingFileSize = unpacked.length();
            }

            if (Math.abs(existingFileSize - size) < 8092) {
                //file is the same as before! Let's continue!
                return unpacked.getPath();
            }

            int chunkSize = 8092;

            //need to read the video in chunks because some old devices can't load it into memory...
            byte[] buffer = new byte[chunkSize];

            outStream = new FileOutputStream(unpacked);

            while (inStream.available() > 0) {
                inStream.read(buffer);
                outStream.write(buffer);
            }
        } catch (Exception e) { 
            Log.e(Extension.EXT_LOG_TAG, "Failed to unpack file" + assetPath);
            return null;
        } finally {
            if (inStream != null) {
                try {
                    inStream.close();
                } catch (Exception e) {
                    Log.e(Extension.EXT_LOG_TAG, "Failed to close input stream for file " + assetPath);
                }
                
            }
            if (outStream != null) {
                try {
                    outStream.flush();
                    outStream.close();
                } catch (Exception e) {
                    Log.e(Extension.EXT_LOG_TAG, "Failed to close output stream for file " + assetPath);
                }
            }
        }

        return unpacked.getPath();
    }

    /*** BINK VIDEO ***/
    public static native int initBinkJNI(AssetManager mgr, boolean useObb, String obbPath);
    public static int initBink() {
        if (loadedStage != 0) {
            return 1;
        }

        int ret =  initBinkJNI(sharedActivity.getAssets(), usingObb, NoodleXBridge.getObbPath());

        loadedStage = 1;
        if(!(nextLoadVideo.equals("")) && !(nextLoadDir.equals("")))
        {
            loadVideo(nextLoadVideo, nextLoadDir);
            nextLoadVideo = "";
            nextLoadDir = "";
        }

        return ret;
    }

    public static native int reinitBinkJNI(AssetManager mgr);
    public static void reinitBink() {
        reinitBinkJNI(sharedActivity.getAssets());
        
        if(!(nextLoadVideo.equals("")) && !(nextLoadDir.equals("")))
        {
            loadVideo(nextLoadVideo, nextLoadDir);
            nextLoadVideo = "";
            nextLoadDir = "";
        }
    }

    public static void setVideoSuffix(String nSuffix) {
        //We don't actually need the video suffix on Android, since we only package one version of the videos in
    }

    public static native int loadVideoJNI(String path, long offset);
    public static int loadVideo(String file, String folder) {
    
        if(loadedStage == 0)
        {
            if(!(nextLoadVideo.equals(""))) {
                 nextLoadVideo = "";
            }
            nextLoadVideo = file;
            return 0;
        }

        String path = "videos/" + file + ".bik";
        long offset = 0;
        
        if (usingObb) {
            offset = NoodleXBridge.getObbOffsetForFile(path);
        }

        int ret = loadVideoJNI(path, offset);

        loadedStage = 3;
        gotoFrame(1);

        return ret;
    }

    public static native int getTotalFramesJNI();
    public static int getTotalFrames() {
        return getTotalFramesJNI();
    }

    public static native boolean nextFrameJNI();
    public static boolean nextFrame() {
        if (loadedStage != 3) {
            return false;
        }
        return nextFrameJNI();
    }

    public static native boolean gotoFrameJNI(int frame);
    public static boolean gotoFrame(int frame) {
        if (loadedStage != 3) {
            return false;
        }
        return gotoFrameJNI(frame);
    }

    public static native void setdrawJNI(double nx0, double ny0, double nx1, double ny1);
    public static void setdraw(double nx0, double ny0, double nx1, double ny1) {
        setdrawJNI(nx0, ny0, nx1, ny1);
    }

    public static native void drawJNI();
    public static void draw() {
        drawJNI();
    }

    public static int loadDualVideo(String file, String folder) {
        //not implemented - doesn't seem to be used
        //Dan said it was an experiment
        return 1;
    }

    public static void toggleDual(int on) {
        //not implemented - doesn't seem to be used
        //Dan said it was an experiment
    }


    /*** VIEWS AND TRANSITIONS ***/
    // This functions adds a loading circle to the opening logo scene, if the OBB needs to be downloaded
    public static void startLoadingObbView ()
    {
        if (obbLoadingImage != null) return;
        
        try
        {
            // Get screensize information
            Display display = sharedActivity.getWindowManager().getDefaultDisplay();
            Point screenSize = new Point();
            display.getSize(screenSize);
           
            // create loading image
            InputStream obbImageStream = sharedActivity.getAssets().open("special/loading.png");
            Drawable obbDrawable = Drawable.createFromStream(obbImageStream, null);
            // Create the loading image view
            obbLoadingImage = new ImageView(sharedActivity);
            obbLoadingImage.setImageDrawable(obbDrawable);
            obbImageStream.close();
            // Create obb loading image parameters
            FrameLayout.LayoutParams obbLoadingParams = new FrameLayout.LayoutParams((int)(screenSize.y*0.25), (int)(screenSize.y*0.25), Gravity.CENTER_HORIZONTAL);
            obbLoadingParams.topMargin = (int)(screenSize.y*0.65);
            // Add to the screen
            sharedActivity.addContentView(obbLoadingImage, obbLoadingParams);
            // Set up the Spin Animation
            RotateAnimation rotAnimation = new RotateAnimation(0, 360, Animation.RELATIVE_TO_SELF, 0.5f, Animation.RELATIVE_TO_SELF, 0.5f);
            rotAnimation.setInterpolator(new LinearInterpolator());
            rotAnimation.setRepeatCount(Animation.INFINITE);
            rotAnimation.setRepeatMode(Animation.INFINITE);
            rotAnimation.setDuration(1000);
            
            // Start the animation
            obbLoadingImage.startAnimation(rotAnimation);
        }
        catch (Exception e)
        {
            Log.e(Extension.EXT_LOG_TAG, "Exception while setting obb loading ImageView: " + e.toString());
        }
    }
    // This method removes the loading obb view, called when obb is finsihed loading.
    public static void stopLoadingObbView ()
    {
        if (obbLoadingImage != null) 
        {
            // Remove the loading image
            obbLoadingImage.clearAnimation();
            obbLoadingImage.setAnimation(null);
            ((ViewGroup) obbLoadingImage.getParent()).removeView(obbLoadingImage);
            obbLoadingImage = null;
        }
    }

    public static void startFirstScreen() 
    {
        try
        {
            // Get the image and assign it to a view
            InputStream imageStream = sharedActivity.getAssets().open("special/luminocitylogo.png");
            Drawable drawable = Drawable.createFromStream(imageStream, null);
            logoImage = new ImageView(sharedActivity);
            logoImage.setImageDrawable(drawable);
            imageStream.close();
            // Determine Image Scalar 
            // NOTE if adding animation CUT THIS:
            Display display = sharedActivity.getWindowManager().getDefaultDisplay();
            Point screenSize = new Point();
            display.getSize(screenSize);
            final float targetScale = 0.6f * ( screenSize.x / screenSize.y);
            logoImage.setScaleX(targetScale);
            logoImage.setScaleY(targetScale);
            // END CUT
            LinearLayout.LayoutParams logoParams = new LinearLayout.LayoutParams(LinearLayout.LayoutParams.MATCH_PARENT, LinearLayout.LayoutParams.MATCH_PARENT, Gravity.CENTER);
            // add view to activity (to see it)
            sharedActivity.addContentView(logoImage, logoParams);

            /*  NOTE decided to NOT include opening animation on android because it was choppy during startup.
            // Prep Image for Animation
            firstScreen.setVisibility(ViewGroup.INVISIBLE);

            // Run the image animation on a separate thread
            sharedActivity.runOnUiThread (new Thread(new Runnable() 
            { 
                public void run() 
                {
                    //Log.d(Extension.EXT_LOG_TAG, "Running start animation on UI thread.");
                    
                    // Get Screen Specs
                    Display display = sharedActivity.getWindowManager().getDefaultDisplay();
                    Point screenSize = new Point();
                    display.getSize(screenSize);

                    // Determine Image Scalar
                    final float targetScale = 0.6f * ( screenSize.x / screenSize.y);

                    // Animation Variables
                    final long startDelay = 1000;     // how long before the animation starts
                    final long scaleDuration = 5000;  // How long the scale animation should take
                    final long fadeDuration = 4000;   // How long the fade in animation should take   

                    // Create Animation Set
                    AnimationSet startingAnimation = new AnimationSet(true);
                    startingAnimation.setInterpolator(new AccelerateDecelerateInterpolator());
                    startingAnimation.setFillEnabled(true);
                    startingAnimation.setFillAfter(true);
                    startingAnimation.setStartOffset(startDelay);
                    startingAnimation.setAnimationListener(new Animation.AnimationListener(){
                        public void onAnimationStart(Animation arg0){
                            // Make Screen Visible
                            firstScreen.setVisibility(ViewGroup.VISIBLE);
                            
                            // Play audio
                            firstAudioPlayer.start();
                        }           
                        public void onAnimationRepeat(Animation arg0){

                        }           
                        public void onAnimationEnd(Animation arg0)
                        {
                            firstScreen.setAlpha(1.0f);

                            // release the audio player
                            firstAudioPlayer.release();
                        }
                    });

                    // Scale Animation
                    Animation scaleAnimation = new ScaleAnimation(  targetScale*0.8f, targetScale, targetScale*0.8f, targetScale, 
                                                                    Animation.RELATIVE_TO_SELF, 0.5f, Animation.RELATIVE_TO_SELF, 0.5f);
                    scaleAnimation.setDuration(scaleDuration);
                    startingAnimation.addAnimation(scaleAnimation);

                    // Alpha Animation
                    Animation alphaAnimation = new AlphaAnimation(0.0f, 1.0f);
                    alphaAnimation.setDuration(fadeDuration);
                    startingAnimation.addAnimation(alphaAnimation);

                    // Play Animation and sounds
                    firstScreen.startAnimation(startingAnimation);
                }
            }));
             */
        }
        catch (Exception e)
        { 
            Log.e(Extension.EXT_LOG_TAG, "Exception while setting image startFirstScreen: " + e.toString());
            return;
        }
       
        // Play intro audio clip
        firstAudioPlayer = new MediaPlayer();
        String cachedAudioFile = unpackAssetToCache("special/intro_to_MENU.mp3", false);
        if (cachedAudioFile != null)
        {
            try
            {
                // Set audio source
                firstAudioPlayer.setDataSource(cachedAudioFile);
                            
                // Prepare and play audio
                firstAudioPlayer.prepare();
                firstAudioPlayer.start();
            }
            catch (Exception e)
            {
                Log.e(Extension.EXT_LOG_TAG, "Error playing firstscreen audio clip: " + e.toString());
            }
            
        }
    }

    public static void stopFirstScreen() 
    {
        // Alpha animation
        sharedActivity.runOnUiThread (new Thread(new Runnable() 
            { 
                public void run() 
                {   
                    logoImage.animate().alpha(0.0f).setDuration(1000).setInterpolator(new AccelerateDecelerateInterpolator())
                        .withEndAction(new Runnable()
                        {
                            // Clean up animations
                            public void run ()
                            {
                                logoImage.clearAnimation();
                                logoImage.setAnimation(null);
                                logoImage.setAlpha(0.0f);

                                // remove loading obb stuff 
                                stopLoadingObbView();
                            }
                        });
                }
            }));
    }

    public static void endFirstScreen() 
    {
        // Cleand up first screen
        ((ViewGroup) logoImage.getParent()).removeView(logoImage);
        logoImage = null;

        firstAudioPlayer.reset();
        firstAudioPlayer = null;
        
        // Check to show creative
        NoodleXBridge.showCreative();
    }

    public static void shouldShowLoadingCircle(boolean value) {
        showLoadingCircle = value;
    }

    public static void addLoadingCircle ()
    {
        try
        {
            // Create the loading screen black background
            if (loadingView == null)
            {
                loadingView = new ImageView(sharedActivity);
                loadingView.setBackgroundColor(Color.BLACK);
                loadingView.setScaleType(ScaleType.FIT_XY);
                LinearLayout.LayoutParams loadingLayout = new LinearLayout.LayoutParams(LinearLayout.LayoutParams.MATCH_PARENT, LinearLayout.LayoutParams.MATCH_PARENT);
                sharedActivity.addContentView(loadingView, loadingLayout);
            }
            loadingView.setVisibility(View.VISIBLE);
            loadingView.bringToFront();
            
            // Create the loading screen icon
            if (loadingImage == null)
            {
                // Get the loading screen image
                InputStream imageStream = sharedActivity.getAssets().open("special/loading.png");
                Drawable drawable = Drawable.createFromStream(imageStream, null);
                imageStream.close();
                // Create the loading image view
                loadingImage = new ImageView(sharedActivity);
                loadingImage.setImageDrawable(drawable);
                LinearLayout.LayoutParams layout = new LinearLayout.LayoutParams(LinearLayout.LayoutParams.MATCH_PARENT, LinearLayout.LayoutParams.MATCH_PARENT);
                layout.gravity = Gravity.CENTER;
                sharedActivity.addContentView(loadingImage, layout); 
                
                // Set up the image
                Display display = sharedActivity.getWindowManager().getDefaultDisplay();
                Point screenSize = new Point();
                display.getSize(screenSize);
                float loadingSize = 0.2f * ( screenSize.x / screenSize.y);
                loadingImage.setScaleX(loadingSize);
                loadingImage.setScaleY(loadingSize); 
            }
            loadingImage.setVisibility(View.VISIBLE);
            loadingImage.bringToFront();
            
            // Set up the Spin Animation
            RotateAnimation rotAnimation = new RotateAnimation(0, 360, Animation.RELATIVE_TO_SELF, 0.5f, Animation.RELATIVE_TO_SELF, 0.5f);
            rotAnimation.setInterpolator(new LinearInterpolator());
            rotAnimation.setRepeatCount(Animation.INFINITE);
            rotAnimation.setRepeatMode(Animation.INFINITE);
            rotAnimation.setDuration(700);

            // Start the Animation
            loadingImage.startAnimation(rotAnimation);
        }
        catch (Exception e)
        {
            Log.e(Extension.EXT_LOG_TAG, "Error setting up loading screen: " + e.toString());
            return;
        }
    }


    public static void setGlobalAlpha(double n) 
    {   
        // Convert new alpha to float
        float newAlpha = (float) n;

        // Check for loadingview stuff
        if (n >= 1) 
        {
            if(loadingView != null) 
            {
                // Remove the loading view
                ((ViewGroup) loadingView.getParent()).removeView(loadingView);
                loadingView = null;
            }
            if (loadingImage != null) 
            {
                // Remove the loading image
                loadingImage.clearAnimation();
                loadingImage.setAnimation(null);
                ((ViewGroup) loadingImage.getParent()).removeView(loadingImage);
                loadingImage = null;
            }
        }
        else if ((logoImage == null) && (loadingView == null) && showLoadingCircle) 
        {
            addLoadingCircle();
        }

        // Do some loading view alpha magic
        if ((loadingView != null) && (loadingImage != null)) 
        {
            loadingView.setAlpha(5f * ((1f-newAlpha) - 0.8f));
            loadingImage.setAlpha(5f * ((1f-newAlpha) - 0.8f));
        }

        // Using Black Layer Hack, can't change the opacity of the adobe air view
        if (fadeView == null)
        {
            // Initialize the fade layer
            fadeView = new ImageView(sharedActivity);
            fadeView.setBackgroundColor(Color.BLACK);
            fadeView.setScaleType(ScaleType.FIT_XY);
            LinearLayout.LayoutParams fadeLayout = new LinearLayout.LayoutParams(LinearLayout.LayoutParams.MATCH_PARENT, LinearLayout.LayoutParams.MATCH_PARENT);
            sharedActivity.addContentView(fadeView, fadeLayout);
            
            // Bring the LuminoLogo image to the front
            if (logoImage != null)
            {
                logoImage.bringToFront();
                if (obbLoadingImage != null)
                {
                    obbLoadingImage.bringToFront();
                }
            }
        }
        fadeView.setAlpha(1.0f-newAlpha);

        // Adjust credits video alpha
        if (videoView != null) 
        {
            videoView.setAlpha(newAlpha);
        }
        if (videoHomeButton != null) 
        {
            videoHomeButton.setAlpha(newAlpha);
        }
    }

    // We don't actually use the uiScale in the Android implementation
    public static void playAV(String name, double uiScale) 
    {
        // load the video
        Uri cachedUri = Uri.fromFile(new File(unpackAssetToCache(name + ".mp4", usingObb)));
        if (cachedUri == null)
        {
            return;
        }

        // Add the video view to the adobe air framelayout
        mainView = (SurfaceView)(sharedActivity.getCurrentFocus());

        // Screen variables
        int screenHeight = sharedActivity.getCurrentFocus().getHeight();
        int screenWidth = sharedActivity.getCurrentFocus().getWidth();
        
        // Setup video view
        if(videoView == null)
        {
            videoView = new VideoView(sharedActivity){
                // Once we add the video view, it steals the back button event
                // Default android behaviour is to exit the application on back, so that's not good
                // Intercept the back and call back into AIR
                public boolean dispatchKeyEvent(KeyEvent event)
                {
                    context.dispatchStatusEventAsync("BACK_PRESSED","TRUE");
                    return true;
                }
            };
            videoView.setVideoURI(cachedUri);

            // Make sure the video will go back to the main menu when it's done playing
            videoView.setOnCompletionListener(new MediaPlayer.OnCompletionListener() {
                @Override
                public void onCompletion(MediaPlayer mp) {
                    context.dispatchStatusEventAsync("BACK_PRESSED","TRUE");
                }
            });
            
            // Scale VideoView for screen size
            MediaMetadataRetriever metaRetriever = new MediaMetadataRetriever();
            metaRetriever.setDataSource(sharedActivity, cachedUri);

            int videoWidth = 1920;
            int videoHeight = 1080;

            String widthString = metaRetriever.extractMetadata(MediaMetadataRetriever.METADATA_KEY_VIDEO_WIDTH);
            if (widthString != null) {
                videoWidth = Integer.parseInt(widthString);
            }

            String heightString = metaRetriever.extractMetadata(MediaMetadataRetriever.METADATA_KEY_VIDEO_HEIGHT);
            if (heightString != null) {
                videoHeight = Integer.parseInt(heightString);
            }
            

            // Set View layout
            FrameLayout.LayoutParams videoLayout = (new FrameLayout.LayoutParams(   ViewGroup.LayoutParams.MATCH_PARENT,
                                                                                    ViewGroup.LayoutParams.MATCH_PARENT));

            float stageScaled = Math.max((float)screenHeight / (float)videoHeight, (float)screenWidth / (float)videoWidth);

            //gotta black bar the sides if the screen is too wide
            if ((float)screenWidth / (float)screenHeight > 16.0 / 9.0) {
                stageScaled = ((float) screenHeight / (float) videoHeight);
            }
                                                                    
            videoLayout.width = (int) (videoWidth * stageScaled);
            videoLayout.height = (int) (videoHeight * stageScaled);
            videoLayout.gravity = Gravity.CENTER;
            videoView.setLayoutParams(videoLayout);
            videoView.setZOrderOnTop(true);
            ((FrameLayout) (mainView.getParent())).addView(videoView);

            // Play video
            videoView.start(); 
        }

        // We used to try and add a menu button in the top right, but due to adobe air surface view nonsense, this would require a
        // huge amount of experimentation to maybe get to work. Instead, the back button can be used on Android to return to the
        // main menu.
    }

    public static void stopAV() 
    {
        // Remove the home button
        if (videoHomeButton != null)
        {
            ((ViewGroup) videoHomeButton.getParent()).removeView(videoHomeButton);
            videoHomeButton = null;
        }
        
        // Remove the video view
        if (videoView != null)
        {
            ((ViewGroup) videoView.getParent()).removeView(videoView);
            videoView = null; 
        }
    }

    public static void setShowVideoHome(boolean show) 
    {
        /*
         * NOTE This hides/shows the Home button during the final credit screen.
         *  Not using due to not being able to show a UI over top of the credits video.
         *  This might not actually work, couldn't check due to not beign able to get both buttons and video working simultaneously.
        /* 
        //Log.d(Extension.EXT_LOG_TAG, "setShowVideoHome start.");
        if (videoHomeButton != null)
        {
            final float fromAlpha;
            final float toAlpha;

            // Initialize from/to values
            if (videoHomeButton.getAlpha() == 0.0f)
            {
                //Log.d(Extension.EXT_LOG_TAG, "fade in!");
                fromAlpha = 0.0f;
                toAlpha = 1.0f;
            }
            else
            {
                //Log.d(Extension.EXT_LOG_TAG, "fade out!");
                fromAlpha = 1.0f;
                toAlpha = 0.0f;
            }
    
            // Animate homebutton
            videoHomeButton.setAlpha(fromAlpha);
            videoHomeButton.animate().alpha(toAlpha).setDuration(500).withEndAction(new Runnable()
            {
                public void run ()
                {
                    //Log.d(Extension.EXT_LOG_TAG, "finished fade in/out.");
                    videoHomeButton.setAlpha(toAlpha);
                }
            });
        }
        */
    }

    public static void restart() 
    {
        // No additional work we have to do on Android to restart after the credits are played
    }


    public static void setImmersiveMode() {
        NoodleXBridge.setImmersiveMode();
        NoodleXBridge.keepScreenOn();
    }

    public static int getProcessMem() {
        return memAvailable;
    }

    public static void requestWriteExternal() {
        NoodleXBridge.requestPermission(NoodleXBridge.NOODLEX_WRITE_EXTERNAL_STORAGE);
    }


/*** GOOGLE PLAY ***/
    public static void authenticateLocalUser() {
        if (didAuthGPGS) return;
        // Authenticate GPGS via NoodleXBridge
        NoodleXBridge.authenticateGPGS();
        didAuthGPGS = true;
    }

    public static void submitAchievement(String id) 
    {
        // Pass id, and cloudload callback into the noodlexgpgs handler
        NoodleXBridge.handleUnlockAchievement(id);
    }

    public static void showAchievements() {
        //Log.d(NoodleXBridge.TAG, "Achievements button pressed.");
        NoodleXBridge.handleViewAchievements();
    }

    // Only rate if the preference does not exist. If it does then the user doesnt want to rate, or already has.
    public static boolean shouldShowRate() 
    {
        // check if the preference exists
        if (preferences.contains(hasRated))
        {
            return false;
        }
        else
        {
            prefEditor.putInt(hasRated, 1);
            prefEditor.apply();
            return true;
        }
    }


/*** MISC ***/
    // This method handles muting/unmuting for when noodlenews creatives are played
    private static boolean mutedForCreative = false;  //  True when the audio was muted for a creative. Unmute on creative dismissed. Prevents unmuting when muted before creative.
    public static void noodlenewsMuteAudio() {
        // Check to see if a creative should mute the game, and that the game is not already muted.
        if (NoodleXBridge.noodlenewsMuteAudio() /* && !MUTED*/)
        {
            // Mute game audio
            context.dispatchStatusEventAsync("PAUSEAUDIO", "TRUE");
            // Flag that game was muted for creative
            mutedForCreative = true;
        }
        else
        {
            // Check to unmute
            if (mutedForCreative)
            {
                // Unmute game Audio
                context.dispatchStatusEventAsync("PAUSEAUDIO", "FALSE");
                // Flag no longer muted for creative
                mutedForCreative = false;
            }
        }
    }


    public static String getPreferredLocalisations() {
        return NoodleXBridge.getLocalizationCode();
    }

    public static void takeScreenshot(double scale) {
        //TODO BEN screenshot
        // float photoScale = scale;
        // if(!whiteFlash) {
        //     whiteFlash = [[UIView alloc] initWithFrame:CGRectMake(0,0,1920,1080)];
        //     whiteFlash.backgroundColor = [UIColor whiteColor];
        //     UIView* standalone = [self findCTStandaloneView:[[UIApplication sharedApplication] keyWindow]];
        //     [standalone addSubview:whiteFlash];
        // }
        // whiteFlash.layer.opacity = 0;
        // CABasicAnimation *fadeOutAnimation = [CABasicAnimation animationWithKeyPath:@"opacity"];
        // fadeOutAnimation.fromValue = [NSNumber numberWithFloat:1.0];
        // fadeOutAnimation.toValue = [NSNumber numberWithFloat:0.0];
        // fadeOutAnimation.additive = NO;
        // fadeOutAnimation.beginTime = CACurrentMediaTime()+0.1;
        // fadeOutAnimation.removedOnCompletion = YES;
        // fadeOutAnimation.beginTime = 0.0;
        // fadeOutAnimation.duration = 0.4;
        // fadeOutAnimation.fillMode = kCAFillModeForwards;
        // [whiteFlash.layer addAnimation:fadeOutAnimation forKey:nil];
        
        // if(screenShotHolder) {
        //     [screenShotHolder removeFromSuperview];
        //     screenShotHolder = NULL;
        // }
        
        // double delayInSeconds = 0.05;
        // dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, delayInSeconds * NSEC_PER_SEC);
        // dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
            
        //     if(lastSavedImage) {
        //         [lastSavedImage release];
        //         lastSavedImage = NULL;
        //     }
            
        //     lastSavedImage = [self drawGlToImage];
        //     if(!lastSavedImage) return;
            
        //     CGRect screenBounds = [[UIScreen mainScreen] bounds];
        //     float scaleExtra = 1;//screenBounds.size.width / lastSavedImage.size.width;
            
        //     screenShotHolder = [[UIImageView alloc] initWithFrame:CGRectMake(0,0,screenBounds.size.width,screenBounds.size.height)];
        //     [screenShotHolder setImage:lastSavedImage];
        //     CTStageView* theStage = [self getStage];
        //     [theStage addSubview:screenShotHolder];
            
        //     CABasicAnimation *animation = [CABasicAnimation animationWithKeyPath:@"transform.scale"];
        //     animation.duration = 1.0f;
        //     animation.timingFunction = [CAMediaTimingFunction functionWithName:kCAMediaTimingFunctionEaseInEaseOut];
        //     animation.fromValue = [NSNumber numberWithFloat:scaleExtra];
        //     animation.toValue = [NSNumber numberWithFloat:photoScale * scaleExtra];
        //     screenShotHolder.transform = CGAffineTransformScale(CGAffineTransformIdentity, photoScale * scaleExtra, photoScale * scaleExtra);
        //     [screenShotHolder.layer addAnimation:animation forKey:nil];
        // });
    }

    public static void cancelScreenshot() {
        //TODO BEN screenshot
        // float duration = 0.5;
        
        // CABasicAnimation *fadeOutAnimation = [CABasicAnimation animationWithKeyPath:@"opacity"];
        // fadeOutAnimation.fromValue = [NSNumber numberWithFloat:1.0];
        // fadeOutAnimation.toValue = [NSNumber numberWithFloat:0.0];
        // fadeOutAnimation.additive = NO;
        // fadeOutAnimation.removedOnCompletion = YES;
        // fadeOutAnimation.beginTime = 0.0;
        // fadeOutAnimation.duration = duration;
        // fadeOutAnimation.fillMode = kCAFillModeForwards;
        
        // screenShotHolder.layer.opacity = 0;
        // if(screenShotHolder) [screenShotHolder.layer addAnimation:fadeOutAnimation forKey:nil];
        
        // if(screenShotHolder){
        //     dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, duration*1.1 * NSEC_PER_SEC);
        //     dispatch_after(popTime, dispatch_get_main_queue(), ^(void){
        //         if(whiteFlash) [whiteFlash removeFromSuperview];
        //         whiteFlash = NULL;
        //         if(screenShotHolder) [screenShotHolder removeFromSuperview];
        //         screenShotHolder = NULL;
        //         if(lastSavedImage) [lastSavedImage release];
        //         lastSavedImage = NULL;
        //     });
        // }
    }

    public static void downloadScreenshot() {
        //TODO BEN screenshot
        // if(!lastSavedImage) return;
        // NSLog(@"downloading screenshot");
        // UIImageWriteToSavedPhotosAlbum(lastSavedImage, self,@selector(thisImage:hasBeenSavedInPhotoAlbumWithError:usingContextInfo:),NULL);
    }

//TODO BEN screenshot
// - (void)thisImage:(UIImage *)image hasBeenSavedInPhotoAlbumWithError:(NSError *)error usingContextInfo:(void*)ctxInfo {
//     dispatch_async(dispatch_get_main_queue(), ^{
//         if (error) {
//             UIAlertView *alert = [[UIAlertView alloc]
//                                   initWithTitle:[error localizedDescription]
//                                   message:[error localizedRecoverySuggestion]
//                                   delegate:nil
//                                   cancelButtonTitle:NSLocalizedString(@"Continue", @"")
//                                   otherButtonTitles:nil];
            
//             [alert show];
//         }else{
//             UIAlertView *alert = [[UIAlertView alloc]
//                                   initWithTitle:NSLocalizedString(@"Saved", @"")
//                                   message:NSLocalizedString(@"SavedBody", @"")
//                                   delegate:nil
//                                   cancelButtonTitle:NSLocalizedString(@"Continue", @"")
//                                   otherButtonTitles:nil];
            
//             [alert show];
//         }
//     });
// }

    public static void shareScreenshot(double x, double y) {
        //TODO BEN screenshot
        // if(!lastSavedImage) return;
        // UIView* standalone = [self findCTStandaloneView:[[UIApplication sharedApplication] keyWindow]];
        // if(!standalone) return;
        
        // UIViewController* standaloneController = [standalone parentViewController];
        // if(!standaloneController) return;
        // NSURL *myWebsite = [NSURL URLWithString:@"http://www.luminocitygame.com/"];
        
        // NSArray *objectsToShare = @[lastSavedImage,NSLocalizedString(@"SOCIAL", @"")];
        
        // UIActivityViewController *activityVC = [[UIActivityViewController alloc] initWithActivityItems:objectsToShare applicationActivities:nil];
        
        // if(!sharePosition) {
        //     sharePosition = [[UIView alloc] initWithFrame:CGRectMake(pos.x, pos.y, 0, 0)];
        //     [standalone addSubview:sharePosition];
        // }
        
        // activityVC.popoverPresentationController.sourceView = sharePosition;
        
        // NSArray *excludeActivities = @[UIActivityTypeAirDrop,
        //                             UIActivityTypeAssignToContact,
        //                             UIActivityTypeSaveToCameraRoll,
        //                             UIActivityTypeAddToReadingList,
        //                             UIActivityTypePostToVimeo,
        //                             UIActivityTypeOpenInIBooks];
        
        // activityVC.excludedActivityTypes = excludeActivities;
        
        
        // [standaloneController presentViewController:activityVC animated:YES completion:nil];
    }

//TODO BEN screenshot
// In Your GL View
// - (UIImage *)drawGlToImage
// {
//     // Draw OpenGL data to an image context
    
//     CTStageView* drawView = [self getStage];
    
    
//     CGRect screenBounds = [[UIScreen mainScreen] bounds];
//     CGFloat screenScale = [[UIScreen mainScreen] scale];
    
//     CGRect finalFrame = CGRectMake(0, 0, screenBounds.size.width * screenScale, screenBounds.size.height * screenScale);
    
//     CGSize canvasSize = finalFrame.size;
//     UIGraphicsBeginImageContext(canvasSize);
    
//     NSLog(@"drawView.frame: %f,%f",finalFrame.size.width,finalFrame.size.height);
    
//     [drawView drawViewHierarchyInRect:finalFrame afterScreenUpdates:YES];
//     //[drawView drawInRect:CGRectMake(0, 0, canvasSize.width, canvasSize.height)];
    
//     UIImage *resultingImage = UIGraphicsGetImageFromCurrentImageContext();
//     UIGraphicsEndImageContext();
    
//     return resultingImage;
// }

    public static void showMoreGames() {
        NoodleXBridge.showMoreGames();
    }


/*** SAVE/LOAD ***/

    // Save the current profile and progress.
    public static boolean setProfilePos(int profile, int chapter) 
    {
        //*
        if(chapter < 0) return false;
        if(chapter < 1) chapter = 1;
        
        // There is no profile, return true <~ probably to create a new one?
        if(profile <= 0) 
        {
            if(currentChapterPosition == chapter) return false;
            currentChapterPosition = chapter;

            return true;
        }
        
        // Find the profile.
        boolean didChange = false;
        if(profile > 0) 
        {
            // Look up the user's profile via the app's Preferences
            // Key's to look for
            String profileChapterKey = chapterKey+profile;  // this preference stores an int value
            String profileUpdatedKey = updatedKey+profile;  // this preference stores a long value

            // Check to see if this profile is saved (key exists)
            if(!preferences.contains(profileChapterKey))
            {
                // Store this user and chapter
                prefEditor.putInt(profileChapterKey, chapter);
                prefEditor.putLong(profileUpdatedKey, System.currentTimeMillis());
                prefEditor.apply();
                // record change
                didChange = true;
            }
            else
            {
                // Get this profile's current chapter
                int currentChapter = preferences.getInt(profileChapterKey,0);

                // Check to see if the chapter given for this profile is further ahead than what is stored
                if (currentChapter == 0 || currentChapter < chapter)
                {
                    // Store this user and chapter
                    prefEditor.putInt(profileChapterKey, chapter);
                    prefEditor.putLong(profileUpdatedKey, System.currentTimeMillis());
                    prefEditor.apply();
                    // record change
                    didChange = true;
                }
            }

            // TODO Cloudsave stuff
            // Update cloud save if change occured
            if (didChange) syncCloudData();
        }
        
        return didChange;
    }

    // Overwrites a profile with given values
    public static void forceProfile(int profile, int chapter) 
    {
        if (profile > 1)
        {
            String profileChapterKey = chapterKey+profile;  // this preference stores an int value.
            String profileUpdatedKey = updatedKey+profile;  // this preference stores a long value.
            String profileDataKey = dataKey+profile;        // this preference stores a string value.

            // Write/overwrite value
            prefEditor.putInt(profileChapterKey, chapter);
            prefEditor.putLong(profileUpdatedKey, System.currentTimeMillis());
            // removed data if stored
            if (preferences.contains(profileDataKey))
            {
                prefEditor.remove(profileDataKey);
            }
            // Apply changes
            prefEditor.apply();
        }

        // Cloud save stuff
        syncCloudData();
    }
    
    // Gets the chapter associated with a given profile, returns 0 if that profile doesn't exist
    public static int getProfileChapter(int profile) 
    {
        if (profile <= 0) return 1;
        int chapter = 0;
        // Look up the user's profile via the app's Preferences
        String profileChapterKey = chapterKey+profile;  // this preference stores an int value.
        
        // Check to see if this profile is saved (key exists)
        if(preferences.contains(profileChapterKey))
        {
            // Return the 
            chapter = preferences.getInt(profileChapterKey, 0);
            return chapter;
        }

        return chapter;
    }

    // Returns the user's preference for showing tips.
    // If there is a preferences, then the user NOT want tips, similar to GPGS sign in.
    public static boolean getProfileManualTip(int profile) 
    {
        if(profile <= 0) return true;

        String profileTipsKey = tipsKey+profile;  // this preference stores a bool value.
        
        // Check if this key exists
        if(!preferences.contains(profileTipsKey)) return true;
        return false;
    }

    // Sets the user's preference for showing tips.
    // Creates a preference, if a preference exists the user does NOT want tips
    public static void setProfileManualTip(int profile, int value) 
    {
        if(profile <= 0) return;

        String profileTipsKey = tipsKey+profile;  // this preference stores a bool value.

        // Add preference
        if(value == 1)
        {
            // Create if it doesnt exist
            if (!preferences.contains(profileTipsKey))
            {
                    prefEditor.putInt(profileTipsKey, value);
                    prefEditor.apply();
            }
        }
        else
        {
            // remove if it exists
            if (preferences.contains(profileTipsKey))
            {
                prefEditor.remove(profileTipsKey);
                prefEditor.apply();
            }
        }

        // Sync with cloud save
        syncCloudData();
    }

    // Sets the current profile to the selected one
    public static void setSelectedProfile(int profile) 
    {
        if(profile <= 0) return;
        String profileChapterKey = chapterKey+profile;  // this preference stores a bool value.

        // Set the selected profile
        prefEditor.putInt(selectedProfile, profile);
        prefEditor.apply();

        // Get this profiles chapter.
        if (preferences.contains(profileChapterKey))
        {
            currentChapterPosition = preferences.getInt(profileChapterKey,0);
        }
    }

    // Get's the selected profile?
    public static int getSelectedProfile() 
    {
        
        int selected = 1;

        // Get the selected profile.
        if (preferences.contains(selectedProfile))
        {
            selected = preferences.getInt(selectedProfile,1);
        }

        return selected;
    }

    // Saves the players progress
    public static void doSave(int profile, int chapter, String data) 
    {
        if (profile > 0)
        {
            // Set up keystores
            String profileChapterKey = chapterKey+profile;  // this preference stores an int value.
            String profileDataKey = dataKey+profile;        // this preference stores a string value.
            String profileUpdatedKey = updatedKey+profile;  // this preference stores a long value.

            // Check if profile exists
            if (preferences.contains(profileChapterKey))
            {
                // Save the data to this profile if the chapter matches what is stored
                if (chapter == preferences.getInt(profileChapterKey,0))
                {
                    prefEditor.putString(profileDataKey, data);
                    prefEditor.putLong(profileUpdatedKey, System.currentTimeMillis());
                    prefEditor.apply();
                }

                // Do cloudsave stuff
                syncCloudData();
            }
        }
    }

    // Returns the data associated with the given profile
    public static String getResumeData(int profile, int chapter) 
    {
        if (profile > 0)
        {   
            // Set up keystores
            String profileChapterKey = chapterKey+profile;  // this preference stores an int value.
            String profileDataKey = dataKey+profile;        // this preference stores a string value.

            // Check if profile save exists
            if (preferences.contains(profileChapterKey))
            {
                String data = preferences.getString(profileDataKey,"");
                return data;
            }
        }

        return null;
    }

    // Remove all profiles
    public static void clearProfiles() 
    {
        // Remove all profiles.
        for (int i = 1; i < 5; i++)
        {
            // Set up keystores
            String profileChapterKey = chapterKey+i;  // this preference stores an int value.
            String profileDataKey = dataKey+i;        // this preference stores a string value.
            String profileUpdatedKey = updatedKey+i;  // this preference stores a long value.

            // Remove the profile chapter save
            if (preferences.contains(profileChapterKey))
            {
                prefEditor.remove(profileChapterKey);
            }
            // Remove the profile data save
            if (preferences.contains(profileDataKey))
            {
                prefEditor.remove(profileDataKey);
            }
            // Remove the profile updated time save
            if (preferences.contains(profileUpdatedKey))
            {
                prefEditor.remove(profileUpdatedKey);
            }
        }
        
        // apply changes
        prefEditor.apply();
    }


    /*** CLOUD SAVE/LOAD ***/
    // This method is passed to NoodleXGPGS as a callable when performing a cloud load.
    // It's called when noodleX finishes loading from the cloud, and simply gets the loaded string from noodlex.
    public static void updateLocalCloudSaveData ()
    {
        // Get NoodleX Cloudsave string.
        mCloudSaveString = NoodleXBridge.getCloudSaveString();      

        // Convert cloud string to meaningful game data
        if (mCloudSaveString == null || mCloudSaveString.equals(""))
        {
            Log.e(NoodleXBridge.TAG, "mCloudSaveString is null. No cloud data retrived to game.");
            return;
        } 
        else
        {
            convertCloudStringToGameSave();
        }        
    }

    // This method converts the cloudsave string to meaningful game data for save comparison.
    // NOTE: THIS IS GAME SPECIFIC.
    
    public static void convertCloudStringToGameSave ()
    {
        /*
         * WRITE CUSTOM GAME CONVERSION CODE BELOW HERE:
         */
        // Cut off the beginning '{' and ending '}' from the raw string.
        String trimmedCloudString = mCloudSaveString.substring(1,mCloudSaveString.length()-1);
       
        // Separate the pais in the string.
        String[] cloudKeyValuePairs = trimmedCloudString.split(", ");
        
        // Add each pair to the new map. Print out
        String[] keyValuePair;
        for (String entry : cloudKeyValuePairs)
        {
            // Get the Key and Value
            keyValuePair = entry.split("=");
            mCloudSaveMap.put(keyValuePair[0].trim(), keyValuePair[1].trim());
        }

        // Sync cloud saves
        syncCloudData();
    }

    // Syncronizes the data found
    public static void syncCloudData ()
    {
        Log.d(NoodleXBridge.TAG, "Syncing Cloud data.");
        // Make sure the cloud save map, and preferences exists
        if (!gpgsCloudSaveComplete)
        {
            Log.e(NoodleXBridge.TAG, "Cloud data not ready yet.");
            return;
        } 

        // Variables
        boolean updateCloud = false;    // Flag for saving to the cloud.
        boolean updateLocal = false;    // flag for updating local preferences

        // Check rate Game popup
        boolean cloudRated = mCloudSaveMap.containsKey(hasRated);
        boolean localRated = preferences.contains(hasRated);
        if (cloudRated && !localRated)
        {
            prefEditor.putInt(hasRated, 1);
            updateLocal = true;
        }
        else if (!cloudRated && localRated)
        {
            updateCloud = true;
        }

        // Check profile selection
        boolean cloudSelected = mCloudSaveMap.containsKey(selectedProfile);
        boolean localSelected = preferences.contains(selectedProfile);
        if (cloudSelected && !localSelected)
        {
            prefEditor.putInt(selectedProfile, Integer.parseInt(mCloudSaveMap.get(selectedProfile)));
            updateLocal = true;
        }
        else if (!cloudSelected && localSelected)
        {
            updateCloud = true;
        }

        // Compare saved profiles
        for (int i = 1; i < 5; i++)
        {
            // Set up possible preference keys for this profile
            String profileChapterKey = chapterKey+i;   // this preference stores an int value.
            String profileDataKey = dataKey+i;         // this preference stores a string value.
            String profileUpdatedKey = updatedKey+i;   // this preference stores a long value.
            String profileTipsKey = tipsKey+i;         // this preference stores a bool value.

            // get the time stamps for each one
            long cloudStamp;
            if (mCloudSaveMap.containsKey(profileUpdatedKey))
            {
                cloudStamp = Long.parseLong(mCloudSaveMap.get(profileUpdatedKey));
            }
            else
            {
               cloudStamp = 0;
            }
            long localStamp = preferences.getLong(profileUpdatedKey,1);                 
            
            // Check for chapter
            boolean cloudChapter = mCloudSaveMap.containsKey(profileChapterKey);
            boolean localChapter = preferences.contains(profileChapterKey);
            // If on the cloud, but not on the device, use cloud
            if (cloudChapter && !localChapter)
            {
                prefEditor.putInt(profileChapterKey, Integer.parseInt(mCloudSaveMap.get(profileChapterKey)));
                updateLocal = true;
            }
            // On the device but not on the cloud, flag cloud for update
            else if (!cloudChapter && localChapter)
            {
                // Flag to update cloud save
                updateCloud = true;
            }
            // on Both, use most recent to resolve.
            else if (cloudChapter && localChapter)
            {
                if (cloudStamp > localStamp)
                {
                    prefEditor.putInt(profileChapterKey, Integer.parseInt(mCloudSaveMap.get(profileChapterKey)));
                    updateLocal = true;
                }
                else
                {
                    updateCloud = true;
                }
            }

            // Check data
            boolean cloudData = mCloudSaveMap.containsKey(profileDataKey);
            boolean localData = preferences.contains(profileDataKey);
            // If on the cloud, but not on the device, use the cloud data
            if (cloudData && !localData)
            {
                prefEditor.putString(profileDataKey, mCloudSaveMap.get(profileDataKey));
                updateLocal = true;
            }
            // Exists locally but not on cloud, flag to update cloud save
            else if (!cloudData && localData)
            {
                updateCloud = true;
            }
            // Exists on both, use time stamp to resolve.
            else if (cloudData && localData)
            {
                if (cloudStamp > localStamp)
                {
                    prefEditor.putString(profileDataKey, mCloudSaveMap.get(profileDataKey));
                    updateLocal = true;
                }
                else
                {
                    updateCloud = true;
                }
            }
            else
            {
                // Doesn't exist on cloud nor local 
            }
            
            // Check show tips preference.
            boolean cloudTips = mCloudSaveMap.containsKey(profileTipsKey);
            boolean localTips = preferences.contains(profileTipsKey);
            // If on the cloud, but not on the device, use the cloud data
            if (cloudTips && !localTips)
            {
                prefEditor.putBoolean(profileTipsKey, Boolean.parseBoolean(mCloudSaveMap.get(profileTipsKey)));
                updateLocal = true;
            }
            // Exists locally but not on cloud, flag to update cloud save
            else if (!cloudTips && localTips)
            {
                updateCloud = true;
            }
            // Exists on both, use time stamp to resolve.
            else if (cloudTips && localTips)
            {
                if (cloudStamp > localStamp)
                {
                    prefEditor.putBoolean(profileTipsKey, Boolean.parseBoolean(mCloudSaveMap.get(profileTipsKey)));
                    updateLocal = true;
                }
                else
                {
                    updateCloud = true;
                }
            }
            else
            {
                // Doesn't exist on cloud nor local 
            }
        }

        // Update local/cloud saves
        if (updateLocal)
        {
            prefEditor.apply();
        }
        if (updateCloud)
        {
            // Create a new string for the game save data
            String newSaveData = "{"; // <-- add opening brace
            // Get current profile
            if (preferences.contains(selectedProfile))
            {
                newSaveData = newSaveData+selectedProfile+"="+preferences.getInt(selectedProfile,0);
            }
            // Get current profile
            if (preferences.contains(hasRated))
            {
                newSaveData = newSaveData+", "+hasRated+"="+preferences.getInt(hasRated,0);
            }
            // Get profile information
            for (int i = 1; i < 5; i++)
            {
                // Set up possible preference keys for this profile
                String profileChapterKey = chapterKey+i;   // this preference stores an int value.
                String profileDataKey = dataKey+i;         // this preference stores a string value.
                String profileUpdatedKey = updatedKey+i;   // this preference stores a long value.
                String profileTipsKey = tipsKey+i;         // this preference stores a bool value.

                // Get current Profiles
                if (preferences.contains(profileChapterKey))
                {
                    newSaveData = newSaveData+", "+profileChapterKey+"="+preferences.getInt(profileChapterKey,0);
                }
                // Get current Data
                if (preferences.contains(profileDataKey))
                {
                    newSaveData = newSaveData+", "+profileDataKey+"="+preferences.getString(profileDataKey,null);
                }
                // Get current UpdatedTime
                if (preferences.contains(profileUpdatedKey))
                {
                    newSaveData = newSaveData+", "+profileUpdatedKey+"="+preferences.getLong(profileUpdatedKey,0);
                }
                // Get current ShowsTips
                if (preferences.contains(profileTipsKey))
                {
                    newSaveData = newSaveData+", "+profileTipsKey+"="+preferences.getBoolean(profileTipsKey,false);
                }
            }
            // Add closing brace
            newSaveData = newSaveData + "}";

            // Save the new game data to the cloud
            NoodleXBridge.saveToCloud(0,newSaveData, false);
        }
    }

        /*** VIDEO RECORDING ***/
    public static boolean isRecordingAvaliable() {
        //not implemented
        //the iOS version uses ReplayKit to make this works
        //Dan said the feature isn't used much and can be skipped for Android
        return false;
    }

    public static void startRecord() {
        //not implemented
        //the iOS version uses ReplayKit to make this works
        //Dan said the feature isn't used much and can be skipped for Android
    }

    public static void stopRecord() {
        //not implemented
        //the iOS version uses ReplayKit to make this works
        //Dan said the feature isn't used much and can be skipped for Android
    }

    public static boolean isCurrentlyRecording() {
        //not implemented
        //the iOS version uses ReplayKit to make this works
        //Dan said the feature isn't used much and can be skipped for Android
        return false;
    }


/*** APPLE TV ***/

    public static void loadAllProfiles() {
        //not implemented
        //believe it was used to get download content on-demand for tvOS based on the save files on the device
    }

    public static void waitingForChapter(int chapter) {
        //not implemented
        //used for on-demand loading of content on tvOS
    }

    public static double isChapterLoaded(int chapter) {
        //used for on-demand loading of content on tvOS
        //always return 2 - it means the chapter is available
        return 2;
    }
    
    public static void requireAllChapters() {
        //not implemented
        //used for on-demand loading of content on tvOS
    }

    public static void focusAdd(String focusName, double y, double x, double w, double h) {
        //not implemented
        //used for tvOS controls
    }
    
    public static void focusClear() {
        //not implemented
        //used for tvOS controls
    }

    public static void setMouseStartedDrag(boolean value) {
        //not implemented
        //used for tvOS controls
    }

    public static boolean checkTVMenuClicked() {
        //not implemented
        //used for tvOS controls
        return false;
    }

    public static void setTVMenuShouldExit(int exit) {
        //not implemented
        //used for tvOS UI hax
    }

    public static void setTVMenuMode(int mode) {
        //not implemented
        //used for tvOS controls
    }

    public static int getTVMenuAction() {
        //not implemented
        //used for tvOS controls

        return 0;
    }


/*** UNUSED ***/
    public static void clearTest() {
        //not implemented
        //only existed for testing purposes
    }

    public static boolean isDemo() {
        //not implemented
        //pretty sure this was used to set a special game mode when the game was on devices in Apple stores
        return false;
    }

}

