/****************************************************************************
 MADE BY LEALAND. BEN IS COOL. 6th, March, 2017

 This class handles Google Play Sign in, and achievement processing.

 DEBUG MODE ~ change the debugMode variable to true/false to log more information. Includes method start's and finishes and more details within.
 
 ### IMPORTANT ACHIEVEMENT INFO ###
    In order for achievements to be read properly:
        1. Create a file called "noodlexachievements.xml"
        2. Put it in the res/raw directory in your android project.
        3. Copy/Paste the generated resources from the Google Play App Management website.
        4. Change the NAMES of the copy/pasted achievements, in the xml file, to match what is called in-game.
            example:    xml = <string name="achievement_chapter_one">CgkIs6fTzdsBEAIQAg</string>
                        game = "chapter one"
                        change = <string name="chapter one">CgkIs6fTzdsBEAIQAg</string>


 ### IMPORTANT CLOUD SAVE INFO ###
    Currently, NoodleX Cloud Save works by storing game save data as a single String. It is the game's responsibility to compile all saved data into a sinlge string,
     and to decypher a loaded cloud save String into meaningful game data.
    These Three functions MUST be implemented into the GAME-SIDE cloudsave functionality. Copy/Paste these template into the game.
    These allows GPGS cloudsave to callback to the game when it has loaded any cloud saved stuff.

    1. Assign the callback method early Game-Side by:
        .. somewhere in the game's init method ...
        // Set GPGS Cloud Load callback
        NoodleXBridge.setCloudLoadCallback(new Callable<Void>()
            {
                public Void call()
                {   
                    updateLocalCloudSaveData(); // <-- method to be passed into NoodleX that will be called once loading has finished. (see 2. bellow)
                    return null;
                }
            });

    2. The Callback Method and Game-Side String variable
        // Game-Side string to store NoodleX cloud data
        private static String mCloudSaveString = "";        // Game-side cloud save string. 

        // This method is passed to NoodleXGPGS as a callable when performing a cloud load.
        // It's called when noodleX finishes loading from the cloud, and simply gets the loaded string from noodlex.
        public static void updateLocalCloudSaveData ()
        {
            // Get NoodleX Cloudsave string.
            mCloudSaveString = NoodleXBridge.getCloudSaveString();      

            // Convert cloud string to meaningful game data
            if (mCloudSaveString == null)
            {
                Log.d(NoodleXBridge.TAG, "CLOUD~ mCloudSaveString is null. No cloud data retrived to game.");
                return;
            } 
            else
            {
                convertCloudStringToGameSave();
            }        
        }

    3. Convert the cloud save string to meaningful game data
        // GAME SPECIFIC CONVERSION VARIABLES:
        private static HashMap<String,String> mCloudSaveMap;

        // This method converts the cloudsave string to meaningful game data for save comparison. NOTE: THIS IS GAME SPECIFIC.
        public static void convertCloudStringToGameSave ()
        {
            //
            // WRITE CUSTOM GAME CONVERSION CODE BELOW HERE:
            //
            ... game code here...
        }

 ****************************************************************************/

package com.noodlecake.NoodleX;

import java.util.*;
import java.util.Set;
import java.util.concurrent.Callable;

import javax.security.auth.callback.Callback;

import android.util.Log;
import android.util.Xml;

import java.io.IOException;
import java.io.StringReader;
import java.io.InputStream;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;

import android.content.res.XmlResourceParser;

import java.lang.ref.WeakReference;
import java.lang.Object;

import java.security.spec.EllipticCurve;

import android.app.Activity;
import android.app.Application;
import android.app.Dialog;
import android.app.AlertDialog;

import android.os.Bundle;
import android.os.AsyncTask;

import android.content.Intent;
import android.content.IntentSender;
import android.content.Context;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.content.res.Resources;

import com.google.android.gms.games.Games;
import com.google.android.gms.games.Games.GamesOptions;
import com.google.android.gms.games.Games.GamesOptions.Builder;
import com.google.android.gms.games.GamesStatusCodes;
import com.google.android.gms.games.GamesActivityResultCodes;
import com.google.android.gms.games.snapshot.Snapshot;
import com.google.android.gms.games.snapshot.Snapshots;
import com.google.android.gms.games.snapshot.SnapshotMetadata;
import com.google.android.gms.games.snapshot.SnapshotMetadataChange;

import com.google.android.gms.drive.Drive;

import com.google.android.gms.common.*;
import com.google.android.gms.common.ConnectionResult;
import com.google.android.gms.common.GooglePlayServicesUtil;
import com.google.android.gms.common.GoogleApiAvailability;
import com.google.android.gms.common.api.*;
import com.google.android.gms.common.api.GoogleApiClient;
import com.google.android.gms.common.api.Api.ApiOptions.HasOptions;
import com.google.android.gms.common.api.GoogleApiClient.ConnectionCallbacks;
import com.google.android.gms.common.api.GoogleApiClient.OnConnectionFailedListener;

import com.noodlecake.NoodleX.NoodleXBridge;

public class NoodleXGPGS
    implements  GoogleApiClient.ConnectionCallbacks,
                GoogleApiClient.OnConnectionFailedListener
{
    /*
     * VARIABLES
     */
    
    // Core stuff
    private static WeakReference<Activity> mActivity;                       // A weak reference to this app's Activity

    //Authentication stuff
    private static Callable<Void> authenticateCompleteCallback;              // Called if the authentication operation completes or fails

    // Bogus CODE identifiers
    private static final int REQUEST_ACHIEVEMENTS = 9002;                   // Bogus code, used to identify achievements callback
    private static final int RC_SIGN_IN = 9001;                             // Bogus code, used for callbacks to this.

    private static GoogleApiClient mGoogleApiClient;                        // Used for obtaining the API client

    private static boolean mResolvingConnectionFailure = false;             // True = currently resolving sign in error, don't sign in again

    public static NoodleXGPGS mInstance;                                    // An instance of this GPGS for callback handling
    
    public static boolean signInForAchievements = false;                    // True = onActivity callback will open achievements (sign in for achievements).

    // Preferences                 
    private static String signInPref = "noodleXAutoLogin";                  // The name if the file to store autologin pref. If this file exists DONT auto log in.
    private static SharedPreferences preferences;                           // The Preferences for this app. Stores auto-login preference
    private static SharedPreferences.Editor prefEditor;                     // Used to edit the shared preferences.
    private static final String RESOLVING_STATE = "resolving_error";        // The preference that is stored when resolving the state


    // Achievements stuff
    private static final String achievementsRecord = "unlockedAchievements";// A string set preference of all unlocked achievements
    private static ArrayList<Achievement> achievements;                     // List of achievement class objects holding the in-game name and Google Play ID number.
    private static final String xmlFileName = "noodlexachievements";        // Name of the achievements XML

    // Cloud save stuff
    private static long mLastCloudSave = 0;                                 // The time since the last cloud save
    private static long mCloudSaveLimit = 6000;                             // The minimum time allowed between cloud saves
    private static boolean mShouldForceSave = false;                        // Whether or not cloudsave should force save, regardless of save timer.
    private static String mCloudSaveStored;                                 // The save data that will be pushed on a Force Save
    private static int resolveAttempts = 5;                                 // How many times a snapshot conflict should be attempted to be resolved.
    private static Snapshot mSnapshot;                                      // The snapshot that will be loaded/saved to.
    private static String mSnapshotKey = "snapshot_";                       // The name of the snapshot that will for every cloud save.
    private static String mCloudSaveString;                                 // The cloud save data in String format.
    private static byte[] mCloudSaveData;                                   // The data from the cloud save.
    private static byte[] mLocalSaveData;                                   // The data from the local save.
    private static Callable<Void> cloudLoadGameCallback;                    // A method that is called when the cloud load finishes.

    /*************************
       GOOGLE PLAY API STUFF 
     *************************/
    // Sets up connection to GooglePlay APIs
    public static void init (WeakReference<Activity> activity)
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Initializing NoodleXGPGS...");
        // Prevent multiple initializations.
        if (mInstance != null) 
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ GPGS already initialized.");
            return;
        }
        
        // Set the activity
        mActivity = activity;
        // Create a singleton class of NoodleGPGS for handling callbacks
        mInstance = new NoodleXGPGS();
        // Set up preferences
        preferences = activity.get().getPreferences(Context.MODE_PRIVATE);
        // Set up preference editor
        prefEditor = preferences.edit();

        // Create the Google Api Client with access to the Play Games services
        mGoogleApiClient = new GoogleApiClient.Builder(mActivity.get())
                    .addConnectionCallbacks(mInstance)
                    .addOnConnectionFailedListener(mInstance)
                    .addApi(Games.API).addScope(Games.SCOPE_GAMES)      // general Google Play Games API
                    .addApi(Drive.API).addScope(Drive.SCOPE_APPFOLDER)  // for Cloud save
                    .setViewForPopups(mActivity.get().getCurrentFocus())
                    .build();

        // Create a list of achievements
        achievements = new ArrayList<Achievement>();
        parseAchievements();
        
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ NoodleXGPGS initialization complete!");
    }

    // Reconnects if we have disconnected in the background
    public static void onResume () 
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ onResume called...");
        if (mResolvingConnectionFailure || mGoogleApiClient == null) return;
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Handling reconnect...");
        if (shouldAutoSignIn() && !mGoogleApiClient.isConnected()) mGoogleApiClient.connect();
    }

        // Closes connection Google Play API
    public static void onStop ()
    {
        NoodleXGPGS.forceCloudSave();
    }

    public static void onPause() 
    {
        NoodleXGPGS.forceCloudSave();
    }

    // Called when app connects to GooglePlay APIs
    public void onConnected (Bundle connectionHint) 
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ API connected...");

        try {
            if (authenticateCompleteCallback != null) {
                authenticateCompleteCallback.call();
            }
        } catch (Exception e) {
            Log.e(NoodleXBridge.TAG, "GPGS~ Failed to call authenitcate callback");
        }

        // Call achievements after signing in, if flaged for it
        if (signInForAchievements)
        {
            // Open achievements 
            signInForAchievements = false;
            NoodleXGPGS.openAchievementsWindow();
        }

        // Submit tracked achievements
        NoodleXGPGS.submitTrackedAchievements();
    }


    // Handles failure to connect to Google Play API
    public void onConnectionFailed (ConnectionResult connectionResult) 
    {
        if (NoodleXBridge.debug)  Log.d(NoodleXBridge.TAG, "GPGS~ Connection failed. Error: "+connectionResult.toString());
        
        // Check if currently resolving connection issues.
        if (mResolvingConnectionFailure)  return;

        // if the sign-in button was clicked or if auto sign-in is enabled, launch the sign-in flow
        if (shouldAutoSignIn()) 
        {
            mResolvingConnectionFailure = true;
        }

        // Resolve connection failure, check if has resolution
        if (connectionResult.hasResolution()) 
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Connection failed. Result has a solution: "+connectionResult.getErrorCode());
            try 
            {
                connectionResult.startResolutionForResult(mActivity.get(), RC_SIGN_IN);
            } 
            catch (IntentSender.SendIntentException e) 
            {
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Autoresolver did not work...: "+e.toString());
                mGoogleApiClient.connect();
                mResolvingConnectionFailure = false;
            }
        } 
        else 
        {
            Log.e(NoodleXBridge.TAG, "GPGS~ Connection failed. Result does not have a solution...");
            try {
                if (authenticateCompleteCallback != null) {
                    authenticateCompleteCallback.call();
                }
            } catch (Exception e) {
                Log.e(NoodleXBridge.TAG, "GPGS~ Failed to call authenitcate callback");
            }

            mResolvingConnectionFailure = false;
        }
    }


    // Called when the app is suspended, minimized?
    public void onConnectionSuspended (int i) 
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Connection suspended");
        // Attempt to reconnect
        if (shouldAutoSignIn()) mGoogleApiClient.connect();
    }


    // Handles the result of connection resolution
    public static void onActivityResult (int requestCode, int resultCode, Intent intent) 
    {
        if (mInstance == null)
        {
            Log.e(NoodleXBridge.TAG, "GPGS~ onActivityResult NULL INSTANCE.");
            return;
        }

        // Receiving Login Failures/Presenting the login screen
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ onActivityResult called...");
        if (requestCode == RC_SIGN_IN) 
        {
            mResolvingConnectionFailure = false;
            if (resultCode == Activity.RESULT_OK) 
            {
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Conecting via onActivityResult");
                NoodleXGPGS.saveAutoSignInPref(true);
                mGoogleApiClient.connect();
            }
            else 
            {
                if (mActivity.get() == null) return;

                String errorMessage = "";
                if (resultCode == Activity.RESULT_CANCELED) 
                {
                    // The player did not wish to sign in. Save preference
                    if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ The Player cancelled the sign in request.");
                    NoodleXGPGS.saveAutoSignInPref(false);
                }
                else
                {
                     Log.e(NoodleXBridge.TAG, "GPGS~ Sign in error: " + resultCode);

                     // Also save the auto sign in pref here
                     // If you don't then we call onResume, and it tries to reconnect indefinitely
                     NoodleXGPGS.saveAutoSignInPref(false);
                }

                try {
                    if (authenticateCompleteCallback != null) {
                        authenticateCompleteCallback.call();
                    }
                } catch (Exception e) {
                    Log.e(NoodleXBridge.TAG, "GPGS~ Failed to call authenitcate callback");
                }
            }
        }

        // Handle Signing out via Achievements
        if (requestCode == REQUEST_ACHIEVEMENTS)
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Achievements window was been requested.");

            if (resultCode == GamesActivityResultCodes.RESULT_RECONNECT_REQUIRED) {
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ The player logged out via Achievements.");

                // disconnect API and save preferences
                NoodleXGPGS.saveAutoSignInPref(false);
                mGoogleApiClient.disconnect();  
            }         
        }

        // Ensure immersive mode is set.
        NoodleXBridge.setImmersiveMode();
    }



    /*********************
       SIGN IN/OUT STUFF
     *********************/

    // Sets a callback for when a user successfully authenticates or an authentication operation fails.
    public static void setAuthenticateCompleteCallback (Callable<Void> callback)
    {   
        authenticateCompleteCallback = callback;
    }

    // Connects the API and authenticates the user, hook from bridge into the game.
    public static void authenticate ()
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Authenticating GPGS...");

		// Safety Check
        if (mGoogleApiClient == null) 
        {
            Log.e(NoodleXBridge.TAG, "GPGS~ Client not initialized...");

            try {
                if (authenticateCompleteCallback != null) {
                    authenticateCompleteCallback.call();
                }
            } catch (Exception e) {
                Log.e(NoodleXBridge.TAG, "GPGS~ Failed to call authenitcate callback");
            }
            return;
        }

        // Check to see if the user should sign in.
        if (shouldAutoSignIn() && !mGoogleApiClient.isConnected()) 
        {
            mGoogleApiClient.connect();
        } else {
            try {
                if (authenticateCompleteCallback != null) {
                    authenticateCompleteCallback.call();
                }
            } catch (Exception e) {
                Log.e(NoodleXBridge.TAG, "GPGS~ Failed to call authenitcate callback");
            }
        }
    }
    
    // Returns true if the user is authenticated, false otherwise
    public static boolean isAuthenticated() {
        if (mGoogleApiClient != null && mGoogleApiClient.isConnected()) {
            return true;
        }
        return false;
    }

    // Checks if the user should Auto-sign in.
    // If the preference exists, then the user should NOT sign in.
    public static boolean shouldAutoSignIn ()
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Checking if should autoSignIn");

        // Check to see if the Preference exists
        if (preferences.contains(signInPref))
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Should NOT auto sign in.");
            return false;  
        } 
        else
        {   
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ SHOULD auto sign in");
            return true;
        } 
    }


    // Takes in a value that determines preference to signin
    // True = should auto-signin, delete preference if it exists.
    // False = should NOT auto signin, create preference if does not exists.
    public static void saveAutoSignInPref (boolean autoSignIn)
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Saving Auto-Signin Pref: "+autoSignIn);

        // Set preference if it needs to be changed.
        if (preferences.contains(signInPref) && autoSignIn) {
            prefEditor.remove(signInPref);
        } 
        else if (!(preferences.contains(signInPref)) && !(autoSignIn)) {
            prefEditor.putInt(signInPref, 1);
        }

        // Store preferences
        prefEditor.apply();
    }


    // Signs the Player in/out and remembers ther preference
    public static void handleSignInSignOut () 
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Sign in/out button pressed...");

        // Check if not signed in, to sign in
        if (mGoogleApiClient != null && !mGoogleApiClient.isConnected()) 
        {     
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Signing in!");
            
            // Save Sign-in preference
            NoodleXGPGS.saveAutoSignInPref(true);
            
            // Sign in.
            mGoogleApiClient.connect();
        }
        // The player explicitly signed out
        else if (mGoogleApiClient != null && mGoogleApiClient.isConnected())
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Signing out!");
            
            // Save Sign-in preference
            NoodleXGPGS.saveAutoSignInPref(false);
            
            // Sign out.
            Games.signOut(mGoogleApiClient);
            mGoogleApiClient.disconnect();
        }
        else // Something else must be wrong
        {
            Log.e(NoodleXBridge.TAG, "GPGS~ Something went wrong with sign in / sign out...");
        }
    }



    /*********************
       ACHIEVEMENTS STUFF
     *********************/
    // Handles requesting to view achievements.
    public static void handleViewAchievements ()
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Handling achievements call.");

        // Check if the user is not connected
        if (!mGoogleApiClient.isConnected())
        {
            // Flag for opening achievemetns once signed in.
            signInForAchievements = true;
            // Sign in.
            NoodleXGPGS.handleSignInSignOut();
        }
        else
        {
            openAchievementsWindow();
        }
    }


    // This method opens the achievements window
    public static void openAchievementsWindow ()
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Opening achievements window.");

        // Bring up Achievements window.
        mActivity.get().startActivityForResult(Games.Achievements.getAchievementsIntent(mGoogleApiClient), REQUEST_ACHIEVEMENTS);
    }


    // Handles an unlock achievement request.
    public static void handleUnlockAchievement (String achievementName)
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Handling unlock achievements.");

        // Check if logged in
        if(mGoogleApiClient != null && !mGoogleApiClient.isConnected()) return;

        // Get Achievement ID
        String id = NoodleXGPGS.GetAchievementID(achievementName);
        if (id == null) {
            Log.e(NoodleXBridge.TAG, "GPGS~ Unable to find achievement ID for name " + achievementName);
            return;
        }

        // Track achievement as unlocked
        NoodleXGPGS.trackUnlockedAchievement(id);

        // Unlock achievement
        Games.setViewForPopups(mGoogleApiClient, mActivity.get().getCurrentFocus());
        Games.Achievements.unlock(mGoogleApiClient,id);
    }
   

    // Handles an increment achievement request.
    public static void handleIncrementAchievement (String achievementName, int increment)
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Handling increment achievement.");

        // Check if logged in
        if(mGoogleApiClient != null && !mGoogleApiClient.isConnected()) return;

        // Get Achievement ID
        String id = NoodleXGPGS.GetAchievementID(achievementName);
        if (id == null) {
            Log.e(NoodleXBridge.TAG, "GPGS~ Unable to find achievement ID for name " + achievementName);
            return;
        }

        // Increment achievement
        Games.setViewForPopups(mGoogleApiClient, mActivity.get().getCurrentFocus());
        Games.Achievements.increment(mGoogleApiClient,id, increment);
    }


    // Keeps record of all unlocked achievements
    public static void trackUnlockedAchievement (String id)
    {
        Set<String> unlocked;
        if (preferences.contains(achievementsRecord))
        {
            unlocked = preferences.getStringSet(achievementsRecord, new HashSet<String>());
        }
        else
        {
            unlocked = new HashSet<String>();
        }
        
        // Add the new achievement id if it doesnt exist
        if (!unlocked.contains(id))
        {
            unlocked.add(id);
        }

        // Update preferences 
        prefEditor.remove(achievementsRecord);
        prefEditor.putStringSet(achievementsRecord, unlocked);
        prefEditor.apply();
    }

    
    // This method unlocks all tracked achievements
    public static void submitTrackedAchievements ()
    {
        if (preferences.contains(achievementsRecord))
        {
            // Get the list of unlocked achievements
            Set<String> unlocked = preferences.getStringSet(achievementsRecord, new HashSet<String>());
            // Resubmit each tracked achievment
            for (String currID : unlocked)
            {
                // Submit this achievement
                Games.Achievements.unlock(mGoogleApiClient,currID);
            }
        }
    }


    /**************************
       ACHIEVEMENTS XML STUFF
     **************************/
    // The Achievement class
    public static class Achievement
    {
        // VARIABLES
        String name;
        String id;

        // Constructor
        public Achievement (String newName, String newId)
        {
            this.name = newName;
            this.id = newId;
        }
    }


    // Gets the achievement ID given the achievement's name
    public static String GetAchievementID (String name)
    {
        // Safety Check
        if (achievements == null) return null;

        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Getting ID for: "+name);

        // go through the list of achievements
        for (int i = 0; i < achievements.size(); i++)
        {
            Achievement currAchievement = achievements.get(i);

            // Check if this is the right one.
            if (new String(currAchievement.name).equals(name))
            {
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Found achievment!");
                return currAchievement.id;
            }
        }
        Log.e(NoodleXBridge.TAG, "GPGS~ Could not find achievement " + name + ". Number of achievements: "+ achievements.size());
        return null;
    }


    // Instantiate the parser.
    public static void parseAchievements ()
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Making xml parser.");
        
        // Parse achievements using xml parsers
        try 
        {
            // Get the achievemetns resources
            Resources resources = mActivity.get().getResources();
            InputStream inStream = resources.openRawResource(resources.getIdentifier(xmlFileName, "raw", mActivity.get().getPackageName()));
            XmlPullParser parser = Xml.newPullParser();
            parser.setInput(inStream, null);
            parser.nextTag();

            // Read the feed! And create a list of achievements.
            parser.require(XmlPullParser.START_TAG, null, "resources");
            while (parser.next() != XmlPullParser.END_TAG)
            {
                if (parser.getEventType() != XmlPullParser.START_TAG) continue;
                
                String name = parser.getName();
                if (name.equals("string")) 
                {   
                    Achievement newAchievement = readAchievement(parser);
                    {
                        if (newAchievement != null)
                        {
                            achievements.add(newAchievement);
                        }   
                    }
                } 
                else 
                {
                    parser.next();
                }
            }
        }
        catch (Exception e)
        {
            Log.e(NoodleXBridge.TAG, "GPGS~ Something went wrong parsing: "+e.toString());
        } 
    }
    

    // Reads the achievement on this line. 
    // Returns a new Achievement holding the information.
    public static Achievement readAchievement (XmlPullParser parser)
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Reading an achievement.");

        String name = "";
        String id = "";

        try
        {
            parser.require(XmlPullParser.START_TAG, null, "string");
            String tag = parser.getName();
            String attributeType = parser.getAttributeValue(null, "name");

            // Check if the tag for this line is the required entry
            if (tag.equals("string"))
            {
                // Get the name of the achievement
                name = parser.getAttributeValue(null, "name");

                // Get the id of the achievement
                if (parser.next() == XmlPullParser.TEXT)
                {
                    id = parser.getText();
                    parser.nextTag();
                }
            }

            // Prevent making 'achievements' for app data stored in the achievements
            if (!(name.equals("app_id")) && !(name.equals("package_name")))
            {
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ New Achievement: "+name+", "+id);
                return new Achievement(name, id);
            } 
            else
            {
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Achievement was app information...");
                return null;
            }
        }
        catch (Exception e)
        {
            Log.e(NoodleXBridge.TAG, "GPGS~ Something went wrong reading achievement: "+e.toString());
        }

        return null;
    }


    /********************
       CLOUD SAVE STUFF
     ********************/
    // This method is invoked from the callable method in-game to return the cloud saved string once, it is finsihed.
    // This is a way to keep cloud load/save encapsulated within NoodleX.
    public static String getCloudSaveString ()
    {
        return mCloudSaveString;
    }
    // Sets the callback method for when the game is done getting the cloud save, or failed to do so.
    public static void setCloudLoadCallback (Callable<Void> callback)
    {   
        cloudLoadGameCallback = callback;
    }
    /*
     * SET CLOUD LOAD DATA
     *  This method takes in a snapshot and sets the cloudload data accordingly.
     *  It then calls the game-side call back to load the data
     */
    public static void setCloudLoadData (Snapshot snapshot)
    {
        try
        {
            // Set the string data for the snapshot
            if (snapshot != null)
            {
                // use the current snapshot
                mCloudSaveData = snapshot.getSnapshotContents().readFully();
                mCloudSaveString = new String(mCloudSaveData, "UTF-8");
            }
            else
            {
                // The snapshot is illegitemate, return an empty string.
                mCloudSaveString = "";
            }
            // Log the data (debugmode)
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ Snapshot data: "+mCloudSaveString);
        }
        catch (Exception e)
        {
            Log.e(NoodleXBridge.TAG, "GPGS~ CLOUD~ Getting snapshot data exception: "+e.toString());
        }
        finally
        {
            try 
            {
                // Call the gameside callback, to retrieve the loaded data
                if (cloudLoadGameCallback != null)
                {
                    cloudLoadGameCallback.call();
                }
                else
                {
                    if (NoodleXBridge.debug) Log.w(NoodleXBridge.TAG, "GPGS~ CLOUD~ Callable not set correctly!");
                } 
            }
            catch (Exception e)
            {
                Log.e(NoodleXBridge.TAG, "GPGS~ CLOUD~ Callback exception: "+e.toString());
            }
        }
    }
    // Load a saved game from the cloud.
    // snapshotName = the name of the game save on the cloud, defaulted to 0 (the first of four save slots).
    // returns a string representing the data stored (note: the game is responsible for cyphering/decyphering what this string means).
    public static void loadFromCloud (Integer snapshotIndex)
    { 
        if (!mGoogleApiClient.isConnected()) return;

        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ loadFromCloud start. index: "+snapshotIndex);

        // VARIABLES
        final String snapshotName = mSnapshotKey+snapshotIndex;

        // Asynchronously load data from the cloud
        AsyncTask<Void, Void, Boolean> loadTask = new AsyncTask<Void, Void, Boolean>()
        {
            // Run loading in teh background
            protected Boolean doInBackground (Void... params)
            {
                // Load the snapshot from Drive
                Snapshots.OpenSnapshotResult loadResult = Games.Snapshots.open(mGoogleApiClient, snapshotName, false, Snapshots.RESOLUTION_POLICY_MOST_RECENTLY_MODIFIED).await();

                // Pass the snapshot into the cloud loading callback.
                NoodleXGPGS.setCloudLoadData(loadResult.getSnapshot());

                return true;
            }
        };

        // run the asyc load process
        loadTask.execute();

        return;
    }
    /*
     * Force Cloud Save
     *  This method saves cloud data regardless of the limiter.
     */ 
    public static void forceCloudSave ()
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ Force saving...");
        // Check if flagged for forcesave
        if (mShouldForceSave)
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ Should force save.");
            // save to the cloud
            NoodleXGPGS.saveToCloud(1, mCloudSaveStored, true);
        }
        else
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ Should not force save.");
        }
    }

     
    /*
     * Save data to the cloud
     *  This method saves an incoming string to the cloud.
     *  saveData = a string of incoming data to be saved (note: the game is responsible for cyphering/decyphering what this string means).
     */ 
    public static void saveToCloud (Integer snapshotIndex, String saveData, boolean forceSave)
    {
        if (!mGoogleApiClient.isConnected()) return;

        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ saveToCloud start, index: "+snapshotIndex+"  data: "+saveData);
        
        // Check cloud save limiter
        if (((System.currentTimeMillis()-mLastCloudSave) < mCloudSaveLimit) && !forceSave)
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ Saving too soon, storing data.");
            
            // Save data and flag for forcesave
            mCloudSaveStored = saveData;
            mShouldForceSave = true;
            return;
        }

        // Set up variables
        final Integer index = snapshotIndex;
        final String data = saveData;

        // Asynchronously save data to the cloud
        AsyncTask<Void, Void, Boolean> saveTask = new AsyncTask<Void, Void, Boolean>()
        {
            // Run saving in background
            protected Boolean doInBackground (Void... params)
            {
                // Check if the snapshot has already been retrieved.
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ Getting snapshot! Creating if needed.");

                // Create the snapshot name
                String snapshotName = mSnapshotKey+index;

                // Open the saved game using its name, creating a new one if it's missing
                Snapshots.OpenSnapshotResult snapshotResult = Games.Snapshots.open(mGoogleApiClient, snapshotName, true, Snapshots.RESOLUTION_POLICY_MOST_RECENTLY_MODIFIED).await();

                // Assign the snapshot
                Snapshot snapshot = snapshotResult.getSnapshot();
                
                // Make sure the resolution was good
                if (snapshot != null)
                {
                    // Write the data to the snapshot
                    snapshot.getSnapshotContents().writeBytes(data.getBytes());

                    // Set metadata
                    SnapshotMetadataChange metadata = new SnapshotMetadataChange.Builder()
                                                                        .setDescription("NoodleX Save")
                                                                        .build();

                    // Commit the change
                    Games.Snapshots.commitAndClose(mGoogleApiClient, snapshot, metadata);

                    // Update cloudsave timestamp, and unflag for forcesave
                    mLastCloudSave = System.currentTimeMillis();           
                    mShouldForceSave = false; 
                    mCloudSaveStored = "";
                    
                    // Return success!
                    return true;
                }
                else
                {
                    if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ Could not resolve opening snapshot for saving...");
                }

                return snapshotResult.getStatus().isSuccess();
            }
        
            // Debug stuff
            protected void onPostExecute (Boolean result) 
            {
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Async background save completed, status: "+result);
            }
        };

        // Run the saving task
        saveTask.execute();
    }
}   // End of class


/*************
   REFERENCE
 *************

    // We decided these functions weren't needed or in certain causes could potentially cause problems
    // Left here in case we change our minds in the future

    // Used to restore the previous instance of the state
    public static void onCreate (Bundle savedInstanceState)
    {
        Log.d(NoodleXBridge.TAG, "GPGS~ onCreate called...");
        if (mInstance == null) return;

        mResolvingConnectionFailure = savedInstanceState != null && savedInstanceState.getBoolean(RESOLVING_STATE, false);
    }
    
    // Used to restore the previous instance of the state
    public static void onSaveInstanceState(Bundle outState)
    {
        Log.d(NoodleXBridge.TAG, "GPGS~ onSaveInstanceState called...");
        if (mInstance == null) return;

        outState.putBoolean(RESOLVING_STATE, mResolvingConnectionFailure);
    }


    /*
     * RECURSIVE RESULT RESOLVER
     *  This method takes in a result and returns a snapshot.
     *  If the result is a conflict it recursively attempts to solve the confilct, returning a resolved snapshot or null.
     *
    public static Snapshot recursiveResultResolver (Snapshots.OpenSnapshotResult snapshotResult, int attempts)
    {
        // Setup
        int statusCode = snapshotResult.getStatus().getStatusCode();
        Snapshot snapshot = null;

        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ Resolving opening snapshot with resultcode:: "+statusCode);
    
        // Check the result of the open operation
        if (statusCode == GamesStatusCodes.STATUS_OK)
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ Successfully loaded snapshot.");
            return snapshotResult.getSnapshot();
        }
        else if (statusCode == GamesStatusCodes.STATUS_SNAPSHOT_CONFLICT)
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ There is a snapshot conflict...");

            // Compare times last modified, take most recent.
            if (snapshotResult.getSnapshot().getMetadata().getLastModifiedTimestamp() 
                < snapshotResult.getConflictingSnapshot().getMetadata().getLastModifiedTimestamp())
            {
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ using the conflicting snapshot.");
                // use the conficting snapshot
                snapshot = snapshotResult.getConflictingSnapshot();
            }
            else
            {
                snapshot = snapshotResult.getSnapshot();
            }
            
            // Tell google which snapshot to use, record result for recursive call.
            Snapshots.OpenSnapshotResult newResult = Games.Snapshots.resolveConflict(mGoogleApiClient, snapshotResult.getConflictId(), snapshot).await();

            // Recursively resolve this conflict
            if (attempts > 0)
            {
                // Still has some attempts remaining. Attempt to resolve new result.
                return recursiveResultResolver(newResult, attempts-1);
            }
            else
            {
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ CLOUD~ Could not resolve snapshot conflicts...");
                return null;
            }
        }
        else
        {
            // Something bad happened :/
            Log.e(NoodleXBridge.TAG, "GPGS~ CLOUD~ Could not load snapshot: "+snapshotResult.getStatus().getStatusCode());
            return null;
        }
    }

*/