/****************************************************************************
 MADE BY LEALAND. BEN IS COOL. 4th, April, 2017
 
 About: 
    This class is meant to get the correct localization for NoodleX messages.
    It reads localized information from a csv file, uses a Key-Value system.

    ### IMPORTANT LOCALIZATION INFO ###
    1. Call the Localization file "noodlexlocalizations.tsv".
    2. Put the localization file into res/raw directory in the android project (same as achievement file).

 ****************************************************************************/

package com.noodlecake.NoodleX;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import java.lang.ref.WeakReference;
import java.security.spec.EllipticCurve;

import java.util.HashMap;
import java.util.Locale;

import android.util.Log;

import android.content.res.Resources;
import android.content.Context;
import android.content.pm.PackageManager;
import android.content.pm.ApplicationInfo;

import android.app.Activity;

import com.noodlecake.BinkANE.functions.doSave;


import com.noodlecake.NoodleX.NoodleXBridge;

public class NoodleXLocalization
{
    /*** VARIABLES ***/ 
    // Core stuff
    private static WeakReference<Activity> mActivity;                               // A weak reference to this app's Activity
    private static Resources mResources;                                            // A link to the resources file.
    // NoodleX-side Language Mapping stuff
    private static String mNoodleXCurrentLocalizationCode;                          // The code for the current localization
    private static String mNoodleXLocalizationKeys[];                                   // An array of all the languages with translations
    private static HashMap<String, HashMap<String, String>> mNoodleXLocalizationMap;// A Hashmap of Hashmaps! <language, <key, message>>
    private static HashMap<String, String> mNoodleXCurrentLocalization;             // The current language set.
    // Game-side Localization stuff
    private static String mGameCurrentLocalizationCode;                             // The code for the current localization
    private static String mGameLocalizationKeys[];                                  // An array of all the languages with translations
    private static HashMap<String, HashMap<String, String>> mGameLocalizationMap;   // A Hashmap of game-side Hashmaps! <language, <key, message>>
    private static HashMap<String, String> mGameCurrentLocalization;                // The current game-side language set.
    // Key Values stuff
    private static String mDefaultLanguage = "en";                                  // Defaults to this language
    private static String mGameName = "The Game!";                                  // the name of the game, suckah!

    public static void init (WeakReference<Activity> activity)
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "GPGS~ Initializing NoodleXLocalization...");

        // Cache the activity
        mActivity = activity;

        // Initialize the hashmaps
        mNoodleXCurrentLocalization = new HashMap<String, String>();
        mNoodleXLocalizationMap = new HashMap<String, HashMap<String, String>>();
        
        // Setup local variables
        mResources = mActivity.get().getResources();

        // Get the name of the game
        try 
        {
            // Get application info stuff
            PackageManager pManager = mActivity.get().getApplication().getApplicationContext().getPackageManager();
            ApplicationInfo appInfo = pManager.getApplicationInfo(mActivity.get().getPackageName(), 0);

            // Set name 
            mGameName = pManager.getApplicationLabel(appInfo).toString();
        } 
        catch (Exception e) 
        {
             Log.d(NoodleXBridge.TAG, "GPGS~ Error getting app info: "+e.toString());
        }

        // Setup Localization HashMap
        try
        {
            // Set up streams
            InputStream inStreamNoodleX = mResources.openRawResource(mResources.getIdentifier("noodlexlocalizations", "raw", mActivity.get().getPackageName()));
            InputStreamReader streamReaderNoodleX = new InputStreamReader(inStreamNoodleX);
            BufferedReader buffReaderNoodleX = new BufferedReader(streamReaderNoodleX);

            // Read through each line of the file.
            String currLine = buffReaderNoodleX.readLine();
            String[] splitLine = currLine.split("\t");
            mNoodleXLocalizationKeys = new String[splitLine.length];
            
            // Set up the hashmaps for each language
            for (int i = 1; i < splitLine.length; i++)
            {
                mNoodleXLocalizationKeys[i] = splitLine[i];
                mNoodleXLocalizationMap.put(splitLine[i], new HashMap<String,String>());
            }        
            
            // Read through each line, adding localizations as needed
            currLine = buffReaderNoodleX.readLine();
            HashMap<String,String> currLanguage;
            while (currLine != null)
            {
                // Split the line by tabs
                splitLine = currLine.split("\t");

                // Set the translation for each language
                for (int i = 1; i < splitLine.length; i++)
                {
                    // Get the Hashmap for this Language
                    currLanguage = mNoodleXLocalizationMap.get(mNoodleXLocalizationKeys[i]);

                    // Store this localization message
                    currLanguage.put(splitLine[0], splitLine[i]);
                }
                
                // Get the line from the file
                currLine = buffReaderNoodleX.readLine();
            }
            // Close the input stream
            inStreamNoodleX.close();

            // Log the Hashmaps
            for (int k = 1; k < mNoodleXLocalizationKeys.length; k++)
            {
                currLanguage = mNoodleXLocalizationMap.get(mNoodleXLocalizationKeys[k]);
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Language: "+mNoodleXLocalizationKeys[k]);
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Current Map: "+currLanguage.toString());
            }

            // Set current localization to device language, defaulting to English if not supported
            setLocalization(getLocalizationCode());
            
        }
        catch(Exception e)
        {
            Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Error setting up NoodleX localization Hashmap: "+e.toString());
        }
    }


    /*
     * GET MESSAGE
     *  This method takes in a Key string, and returns the coresponding message givent he current localization.
     */
    public static String getMessage (String key)
    {
        String message;
        
        // Get the corresponding message for this key
        if (mNoodleXCurrentLocalization.containsKey(key))
        {
            message = mNoodleXCurrentLocalization.get(key);

            // Check for <GAME>, replace with the name of the game
            message = message.replaceAll("<GAME>", mGameName);
        }
        else if ((mGameCurrentLocalization != null) && (mGameCurrentLocalization.containsKey(key)))
        {
            message = mGameCurrentLocalization.get(key);
        }
        else
        {
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Invalid message key: "+key);
            message = "";
        }

        // Return the message.
        return message;
    }


    /*
     * SET LOCALIZATION
     *  This method takes in a string for a language score
     */
    public static void setLocalization (String language)
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ setLocalization start: "+language);
        /*** Variables ***/
        // String stuff
        String splitLanguage[]; // The recieved language, split into language and dialect
        String languageCode;    // the language component of splitLang
        // Get language and dialect codes
        splitLanguage = language.split("-");
        languageCode = splitLanguage[0];

        /*** SET UP NOODLEX LOCALIZATION ***/
        // Check to see if there is localization for this language.
        if (mNoodleXLocalizationMap.containsKey(language))                // <-- Has localization
        {
            // Set currentLanguage to this language
            mNoodleXCurrentLocalization = mNoodleXLocalizationMap.get(language);
            mNoodleXCurrentLocalizationCode = language;
            
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Setting localization: "+language);
        }
        else if (mNoodleXLocalizationMap.containsKey(languageCode))       // <-- Has localization, but not for this dialect
        {
            // Set currentLanguage to this language, ignoring dialect
            mNoodleXCurrentLocalization = mNoodleXLocalizationMap.get(languageCode);
            mNoodleXCurrentLocalizationCode = languageCode;
            
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Setting localization, no find dialect: "+languageCode);
        }
        else                                                        // <-- No base language, look for alt dialect, or default to english.
        {
            // For for an alternative dialect
            String splitKey[];
            boolean alternativeFound = false;
            for (int i = 1; i < mNoodleXLocalizationKeys.length; i++)
            {
                // Check to see if this is an alternative dialect. Ignoring Chinese
                splitKey = mNoodleXLocalizationKeys[i].split("-");
                if ((!languageCode.equals("zh")) && splitKey[0].equals(languageCode))
                {
                    if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Using alternative dialect: "+mNoodleXLocalizationKeys[i]);
                    
                    // Use this language!
                    mNoodleXCurrentLocalization = mNoodleXLocalizationMap.get(mNoodleXLocalizationKeys[i]);
                    mNoodleXCurrentLocalizationCode = mNoodleXLocalizationKeys[i];
                    break;
                }
            }

            // Default the currentLanguage to english.
            if (!alternativeFound)
            {   
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Setting localization to Default: "+mDefaultLanguage);
                mNoodleXCurrentLocalization = mNoodleXLocalizationMap.get(mDefaultLanguage); 
                mNoodleXCurrentLocalizationCode = mDefaultLanguage;
            }
        }

        /*** SET UP NOODLEX LOCALIZATION ***/
        if (mGameLocalizationMap == null || mGameLocalizationKeys == null) return;
        // Check to see if there is localization for this language.
        if (mGameLocalizationMap.containsKey(language))                // <-- Has localization
        {
            // Set currentLanguage to this language
            mNoodleXCurrentLocalization = mGameLocalizationMap.get(language);
            mGameCurrentLocalizationCode = language;
            
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ GS~ Setting localization: "+language);
        }
        else if (mGameLocalizationMap.containsKey(languageCode))       // <-- Has localization, but not for this dialect
        {
            // Set currentLanguage to this language, ignoring dialect
            mGameCurrentLocalization = mGameLocalizationMap.get(languageCode);
            mGameCurrentLocalizationCode = languageCode;
            
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ GS~ Setting localization, no find dialect: "+languageCode);
        }
        else                                                        // <-- No base language, look for alt dialect, or default to english.
        {
            // For for an alternative dialect
            String splitKey[];
            boolean alternativeFound = false;
            for (int i = 1; i < mGameLocalizationKeys.length; i++)
            {
                // Check to see if this is an alternative dialect. Ignoring Chinese
                splitKey = mGameLocalizationKeys[i].split("-");
                if ((languageCode.equals("zh")) && splitKey[0].equals(languageCode))
                {
                    if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ GS~ Using alternative dialect: "+mGameLocalizationKeys[i]);
                    
                    // Use this language!
                    mGameCurrentLocalization = mGameLocalizationMap.get(mGameLocalizationKeys[i]);
                    mGameCurrentLocalizationCode = mGameLocalizationKeys[i];
                    break;
                }
            }

            // Default the currentLanguage to english.
            if (!alternativeFound)
            {   
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ GS~ Setting localization to Default: "+mDefaultLanguage);
                mGameCurrentLocalization = mGameLocalizationMap.get(mDefaultLanguage); 
                mGameCurrentLocalizationCode = mDefaultLanguage;
            }
        }
    }


    /*
     * GET LOCALIZATIONS
     *  This method returns the code for the current localization
     */
    public static String getLocalizations ()
    {
        return "NoodleX: "+mNoodleXCurrentLocalizationCode+"  Game: "+mGameCurrentLocalizationCode;
    }
    public static String getNoodleXLocalization ()
    {
        return mNoodleXCurrentLocalizationCode;
    }
    public static String getGameLocalization ()
    {
        return mGameCurrentLocalizationCode;
    }


    public static String getLanguageCode() {
        String code = Locale.getDefault().getLanguage();
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Device language: " + code);
        return code;
    }

    public static String getCountryCode() {
        String code = Locale.getDefault().getCountry();
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Country code: " + code);
        return code;
    }

    public static String getLocalizationCode() {
        String code = NoodleXLocalization.getLanguageCode() + "-" + NoodleXLocalization.getCountryCode();
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Localization code: " + code);
        return code;
    }




/*********************************
   GAME-SIDE LOCALIZATION STUFF
 *********************************/
    /*
     * GET LOCALIZATION
     *  This method returns the code for the current localization
     */
    public static void setGameLocalizations (String fileName)
    {
        // Setup Localization HashMap
        try
        {
            // Setup Game-Side hashmaps
            mGameCurrentLocalization = new HashMap<String, String>();
            mGameLocalizationMap = new HashMap<String, HashMap<String, String>>();
            
            // Set up streams
            InputStream inStreamGame = mResources.openRawResource(mResources.getIdentifier(fileName, "raw", mActivity.get().getPackageName()));
            InputStreamReader streamReaderGame = new InputStreamReader(inStreamGame);
            BufferedReader buffReaderGame = new BufferedReader(streamReaderGame);

            // Read through each line of the file.
            String currLine = buffReaderGame.readLine();
            String[] splitLine = currLine.split("\t");
            mGameLocalizationKeys = new String[splitLine.length];
            
            // Set up the hashmaps for each language
            for (int i = 1; i < splitLine.length; i++)
            {
                mGameLocalizationKeys[i] = splitLine[i];
                mGameLocalizationMap.put(splitLine[i], new HashMap<String,String>());
            }        
            
            // Read through each line, adding localizations as needed
            currLine = buffReaderGame.readLine();
            HashMap<String,String> currLanguage;
            while (currLine != null)
            {
                // Split the line by tabs
                splitLine = currLine.split("\t");

                // Set the translation for each language
                for (int i = 1; i < splitLine.length; i++)
                {
                    // Get the Hashmap for this Language
                    currLanguage = mGameLocalizationMap.get(mGameLocalizationKeys[i]);

                    // Store this localization message
                    currLanguage.put(splitLine[0], splitLine[i]);
                }
                
                // Get the line from the file
                currLine = buffReaderGame.readLine();
            }
            // Close the input stream
            inStreamGame.close();

            // Log the Hashmaps
            for (int k = 1; k < mGameLocalizationKeys.length; k++)
            {
                currLanguage = mGameLocalizationMap.get(mGameLocalizationKeys[k]);
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ GS~ Language: "+mGameLocalizationKeys[k]);
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "LOCALIZATION~ GS~ Current Map: "+currLanguage.toString());
            }

            // Set current localization to english
            mGameCurrentLocalization = mGameLocalizationMap.get(mDefaultLanguage);
        }
        catch(Exception e)
        {
            Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Error setting up Game-Side Localization Hashmap: "+e.toString());
        }
    }



/*****************
   TESTING STUFF
 *****************/
    public static void testSetLocalization ()
    {
        // Localization testing!
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ set to existing language: fr");
        NoodleXLocalization.setLocalization("fr");  // existing language
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get localization: fr");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getLocalizations());

        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ set to language without correct dialect: ru-KO");
        NoodleXLocalization.setLocalization("ru-KO");  // base, but no dialect
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get localization: ru");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getLocalizations());
    
        //Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ set to language without default base, but other dialect: zh-YT");
        //NoodleXLocalization.setLocalization("zh-YT");  // no base, but dialect

        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ set to language non existant chinese dialect, default to english: zh-YT");
        NoodleXLocalization.setLocalization("zh-YT");  // Chinese test
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get localization: en");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getLocalizations());

        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ set to non existing dialect nor base: lo");
        NoodleXLocalization.setLocalization("lo");  // non existent
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get localization: en");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getLocalizations());
    }
    public static void testGetMessage ()
    {
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ Set language to English: en");
        NoodleXLocalization.setLocalization("en"); 
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message for existing key: NC_NoInternet");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("NC_NoInternet"));
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message containing <GAME> key: NC_PermissionRequiredRun");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("NC_PermissionRequiredRun"));
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message with invalid key: Derp");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("Derp"));

        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ Set language to Korean: ko");
        NoodleXLocalization.setLocalization("ko"); 
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message for existing key: NoInternet");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("NC_NoInternet"));
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message containing <GAME> key: NC_PermissionRequiredRun");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("NC_PermissionRequiredRun"));
    }

    public static void testSetGameLocalization ()
    {
        // Set Game Localization
        NoodleXLocalization.setGameLocalizations("testlocalizations");

        // Localization testing!
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ set to existing in both game and noodlex: fr");
        NoodleXLocalization.setLocalization("fr");  // existing language
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get localization: fr and fr");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getLocalizations());
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message for existing key in NoodleX: NC_NoInternet");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("NC_NoInternet"));
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message for existing key in Game: TestLerp");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("TestLerp"));
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message for existing key in Game does not exist: TestGerp");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("TestGerp"));

        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ set to language without correct dialect for NoodleX but with Game: ru-KO");
        NoodleXLocalization.setLocalization("ru-KO");  // base, but no dialect
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get localization: ru and ru-KO");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getLocalizations());
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message for existing key in NoodleX: NC_PermissionRequiredRun");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("NC_PermissionRequiredRun"));
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message for existing key in Game: TestDerp");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("TestDerp"));

        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ set to Chinese dialect that in NoodleX, not in Game: zh-TW");
        NoodleXLocalization.setLocalization("zh-TW");  // base, but no dialect
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get localization: zh-TW and en");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getLocalizations());
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message for existing key in NoodleX: NC_NoInternet");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("NC_NoInternet"));
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ TEST~ get message for existing key in Game that exist: TestHerp");
        Log.d(NoodleXBridge.TAG, NoodleXLocalization.getMessage("TestHerp"));
    }

}


/*************
   WORKSPACE
 *************



// Read through each split
for (int i = 0; i < splitLine.length; i++)
{
    String lang;
    String dialect;
    String currLang[];

    // Split the current language into language and dialect
    currLang = splitLine[i].split("-");

    // Set the language
    lang = currLang[0];
    // Set the dialect
    if (currLang.length > 1)
    {
        dialect = currLang[1];
        // Print out line!
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Language: "+lang+"  Dialect: "+dialect);
    }
    else
    {
        // Print out line!
        Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Language: "+lang);
    }
}

        // Buffer/Stream stuff
        String currLine;        // The current line in the file stream
        String[] splitLine;     // The current line split
        String currLang;        // the current splitLine lang component
        String currDial;        // the current splitLine dialect component
        int languageIndex;      // the index which the base language is
        int dialectIndex;      // the index which the dialect specific language is

        // Set the language
        mLanguage = language;

        // Determine the language and dialect of language to be set.
        splitLang = mLanguage.split("-");
        lang = splitLang[0];
        if (splitLang.length > 1) dialect = splitLang[1];

        /*
         * SETUP SPECIFIED LANGUAGE HASHMAP!
         *
        try
        {
            // Set up streams
            InputStream inStream = resources.openRawResource(resources.getIdentifier("noodlexlocalizations", "raw", mActivity.get().getPackageName()));
            InputStreamReader streamReader = new InputStreamReader(inStream);
            BufferedReader buffReader = new BufferedReader(streamReader);

            // Read through each line of the file.
            currLine = buffReader.readLine();
            languageIndex = 0;
            dialectIndex = 0;

            // Look through the first line to find the language
            splitLine = currLine.split("\t");
            for (int i = 0; i < splitLine.length; i++)
            {
                // Determine the language and dialect
                splitLang = splitLine[i].split("-");
                currLang = splitLang[0];
                if (splitLang.length > 1) currDial = splitLang[1];

                // See if it's the correct language and dialect
                if ((lang != null) && (dialect == null))        // <--- there is not dialect for this language
                {
                    if (lang.compareTo(currLang))
                    {

                    } 

                    if (splitLang.length > 1)
                    {

                    }
                }


                else if ((lang != null) && (dialect != null))   // <--- there is a dialect for this language
                {
                    if (splitLang.length > 1)
                    {

                    }
                }

            }



            while (currLine != null)
            {
                // Split the line by tabs
                splitLine = currLine.split("\t");

                // Look through the first line to find the language

                // Get the line from the file
                currLine = buffReader.readLine();
            } 

            // Close the input stream
            inStream.close();
        }
        catch(Exception e)
        {
            Log.d(NoodleXBridge.TAG, "LOCALIZATION~ Error setting up DefaultLanguage Hashmap: "+e.toString());
        }

 */