/****************************************************************************
 MADE BY BEN. BEN IS COOL. Jan 5, 2017
 ****************************************************************************/

package com.noodlecake.NoodleX;

import java.lang.ref.WeakReference;
import android.app.Activity;

import android.os.Environment;
import android.os.Build;
import android.view.View;
import android.view.Gravity;
import android.view.WindowManager;
import android.widget.Space;
import android.graphics.Bitmap;
import android.net.Uri;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.widget.Button;
import android.widget.LinearLayout;
import android.widget.Toast;
import android.app.AlertDialog;
import android.R;
import android.Manifest;
import android.support.v4.content.ContextCompat;
import android.support.v4.app.ActivityCompat;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.content.Context;
import android.app.Activity;

import com.google.android.gms.ads.identifier.AdvertisingIdClient;

import android.util.Log;
import java.lang.*;
import java.util.*;
import java.util.concurrent.Callable;

import com.noodlecake.NoodleX.NoodleXBridge;

public class NoodleXNative {

    private static WeakReference<Activity> mActivity;
    private static int pendingPermission;
    private static String advertisingID;

    private static final String launchCount = "NumberOfLaunches";       // Stores the number of times the app was launched
    private static SharedPreferences preferences;                       // The Preferences for this app. Stores relevant settings
    private static SharedPreferences.Editor prefEditor;                 // Used to edit the shared preferences. 

    private static Callable<Void> permissionCompleteCallback;

    public static void init(WeakReference<Activity> activity){
        Log.d(NoodleXBridge.TAG, "NATIVE~ Initializing NoodleXNative...");

    	mActivity = activity;
        advertisingID = "unknown";
        pendingPermission = NoodleXBridge.NO_PERMISSION;

        // Set up Preferences
        preferences = activity.get().getPreferences(Context.MODE_PRIVATE);
        // Set up preference editor
        prefEditor = preferences.edit();
        // Increment startup
        incrementLaunchCounter();

        //getAdevertisingIdInfo will throw an exception if called from the main thread.
        Thread t = new Thread(new Runnable() {
             public void run() {
                //get Advertising ID - other UUIDs need weird permissions
                String id = "unknown";

                try {
                    AdvertisingIdClient.Info info = AdvertisingIdClient.getAdvertisingIdInfo(mActivity.get());
                    if (info.isLimitAdTrackingEnabled()) {
                        id = info.getId();
                    } else {
                        id = "opt-out";
                    }
                } catch (Exception e) {
                    Log.e(NoodleXBridge.TAG, "NATIVE~ Failed to obtain advertising identifier!");
                }

                advertisingID = id;
             }
        });  
        t.start();

        Log.d(NoodleXBridge.TAG, "NATIVE~ NoodleXNative initialization complete!");
    }

    public static void onResume() {
        if (pendingPermission != NoodleXBridge.NO_PERMISSION) {
            pendingPermission = NoodleXBridge.NO_PERMISSION;

            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Back from asking for permission!");

            if (permissionCompleteCallback != null) {
                try {
                    permissionCompleteCallback.call();
                } catch (Exception e) {
                    Log.e(NoodleXBridge.TAG, "NATIVE~ Permission complete callback failed: " + e.toString());
                }
            }
        }
        onWindowFocusChanged(true);
    }

    public static void onWindowFocusChanged(boolean hasFocus) {
        setImmersiveMode();
        keepScreenOn();
    }

    // This method updates the launch counter, called every time the game launches
    private static void incrementLaunchCounter ()
    {
       if (preferences == null)
        {
            Log.d(NoodleXBridge.TAG, "NATIVE~ Cannot show increment launch counter, preferences not initialized..");
            return;
        }

        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Incrementing launch counter");
        int currCount = preferences.getInt(launchCount, 0);
        prefEditor.putInt(launchCount,currCount+1);
        prefEditor.apply();
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Number of launches: "+preferences.getInt(launchCount,0));
    }

    public static int getLaunchCount() {
        return preferences.getInt(launchCount,0);
    }

    public static void shareText(String message) {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Sharing text...");
        Intent sharingIntent = new Intent(Intent.ACTION_SEND);
        sharingIntent.setType("text/plain");
        sharingIntent.putExtra(Intent.EXTRA_TEXT, message);
        mActivity.get().startActivity(Intent.createChooser(sharingIntent,"Share..."));
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Share intent started.");
    }

    public static void shareImageWithText(String imgPath, String message) {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Sharing image...");

        Intent sharingIntent = new Intent(Intent.ACTION_SEND);
        Uri screenshotUri;
        if (!imgPath.startsWith("file://")) {
            screenshotUri = Uri.parse("file://" + imgPath);
        } else {
            screenshotUri = Uri.parse(imgPath);
        } 

        sharingIntent.setType("image/png");
        sharingIntent.putExtra(Intent.EXTRA_STREAM, screenshotUri);

        if (!message.equals("")) {
            sharingIntent.putExtra(Intent.EXTRA_TEXT, message);
        }

        mActivity.get().startActivity(Intent.createChooser(sharingIntent, "Share..."));

        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Share intent started.");
    }

    public static String getPhotosDirectory() {
        String directory = Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_PICTURES).getAbsolutePath();
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ External photos directory is " + directory);
        return directory;
    }

    public static void openURL(String url) {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Opening URL " + url + "...");

        Uri addressUri = Uri.parse(url);
            
        Intent webIntent = new Intent(Intent.ACTION_VIEW);
        webIntent.setData(addressUri);
        mActivity.get().startActivity(webIntent);

        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Web intent started.");
    }

    public static String getVersionNumber() {
        try {
            PackageInfo info = mActivity.get().getPackageManager().getPackageInfo(mActivity.get().getPackageName(), 0);
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Version number: " + info.versionName);
            return info.versionName;
        } catch (Exception e) {
            return "";
        }
    }

    public static int getVersionCode() {
        try {
            PackageInfo info = mActivity.get().getPackageManager().getPackageInfo(mActivity.get().getPackageName(), 0);
            if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Version code: " + info.versionCode);
            return info.versionCode;
        } catch (Exception e) {
            return 1;
        }
    }

    public static void showOneButtonPopup(String title, String message, String button1Text, final Callable<Void> button1Callback) {

        final String uiTitle = new String(title);
        final String uiMessage = new String(message);
        final String uiButton1Text = new String(button1Text);

        mActivity.get().runOnUiThread (new Thread(new Runnable() { 
             public void run() {
                new AlertDialog.Builder(mActivity.get(), AlertDialog.THEME_DEVICE_DEFAULT_DARK)
                    .setTitle(uiTitle)
                    .setMessage(uiMessage)
                    .setPositiveButton(uiButton1Text, new DialogInterface.OnClickListener() {
                        public void onClick(DialogInterface dialog, int which) { 
                            if (button1Callback != null) {
                                try {
                                    button1Callback.call();
                                } catch (Exception e) {
                                    Log.e(NoodleXBridge.TAG, "NATIVE~ Error in button1callback: " + e.toString());
                                }
                            }
                        }
                     })
                     .show();    
             }
         }));
    }

    public static void showTwoButtonPopup(String title, String message, 
    String button1Text, String button2Text, 
    final Callable<Void> button1Callback, final Callable<Void> button2Callback) {

        final String uiTitle = new String(title);
        final String uiMessage = new String(message);
        final String uiButton1Text = new String(button1Text);
        final String uiButton2Text = new String(button2Text);

        mActivity.get().runOnUiThread (new Thread(new Runnable() { 
             public void run() {
                new AlertDialog.Builder(mActivity.get(), AlertDialog.THEME_DEVICE_DEFAULT_DARK)
                    .setTitle(uiTitle)
                    .setMessage(uiMessage)
                    .setPositiveButton(uiButton1Text, new DialogInterface.OnClickListener() {
                        public void onClick(DialogInterface dialog, int which) { 
                            if (button1Callback != null) {
                                try {
                                    button1Callback.call();
                                } catch (Exception e) {
                                    Log.e(NoodleXBridge.TAG, "NATIVE~ Error in button1callback: " + e.toString());
                                }
                            }
                        }
                     })
                    .setNegativeButton(uiButton2Text, new DialogInterface.OnClickListener() {
                        public void onClick(DialogInterface dialog, int which) { 
                            if (button2Callback != null) {
                                try {
                                    button2Callback.call();
                                } catch (Exception e) {
                                    Log.e(NoodleXBridge.TAG, "NATIVE~ Error in button2callback: " + e.toString());
                                }
                            }
                        }
                     })
                     .show();    
             }
         }));
    }

    public static void showThreeButtonPopup(String title, String message, 
    String button1Text, String button2Text, String button3Text, 
    final Callable<Void> button1Callback, final Callable<Void> button2Callback, final Callable<Void> button3Callback) {

        final String uiTitle = new String(title);
        final String uiMessage = new String(message);
        final String uiButton1Text = new String(button1Text);
        final String uiButton2Text = new String(button2Text);
        final String uiButton3Text = new String(button3Text);

        mActivity.get().runOnUiThread (new Thread(new Runnable() { 
             public void run() {
                final AlertDialog dialog = new AlertDialog.Builder(mActivity.get(), AlertDialog.THEME_DEVICE_DEFAULT_DARK)
                    .setTitle(uiTitle)
                    .setMessage(uiMessage)
                    .setPositiveButton(uiButton1Text, new DialogInterface.OnClickListener() {
                        public void onClick(DialogInterface dialog, int which) { 
                            if (button1Callback != null) {
                                try {
                                    button1Callback.call();
                                } catch (Exception e) {
                                    Log.e(NoodleXBridge.TAG, "NATIVE~ Error in button1callback: " + e.toString());
                                }
                            }
                        }
                     })
                    .setNeutralButton(uiButton2Text, new DialogInterface.OnClickListener() {
                        public void onClick(DialogInterface dialog, int which) { 
                            if (button2Callback != null) {
                                try {
                                    button2Callback.call();
                                } catch (Exception e) {
                                    Log.e(NoodleXBridge.TAG, "NATIVE~ Error in button2callback: " + e.toString());
                                }
                            }
                        }
                     })
                    .setNegativeButton(uiButton3Text, new DialogInterface.OnClickListener() {
                        public void onClick(DialogInterface dialog, int which) { 
                            if (button3Callback != null) {
                                try {
                                    button3Callback.call();
                                } catch (Exception e) {
                                    Log.e(NoodleXBridge.TAG, "NATIVE~ Error in button3callback: " + e.toString());
                                }
                            }
                        }
                     })
                    .create();
                      
                dialog.show();

                try {
                    dialog.setOnShowListener(new DialogInterface.OnShowListener() {
                        @Override
                        public void onShow(DialogInterface dialogInterface) {
                            NoodleXNative.verifyButtonWidths(dialog.getButton(AlertDialog.BUTTON_POSITIVE));
                        }
                    });

                    final Button button = dialog.getButton(AlertDialog.BUTTON_POSITIVE);
                    LinearLayout layout = (LinearLayout) button.getParent();
                    layout.setOrientation(LinearLayout.VERTICAL);
                } catch (Exception e) {
                    //don't worry about it...
                }
             }
         }));
    }

    public static void showToast(String message) {
        // Create a toast
        // it has to run on the UI thread otherwise we crash...

        final String str = message;
        final Activity act = mActivity.get();
        act.runOnUiThread(new Runnable() {
            @Override
            public void run() {
                Toast.makeText(act, str, Toast.LENGTH_LONG).show();
            }
        });
    }

    
    public static void verifyButtonWidths(Button dialogButton) {
        if (dialogButton == null)
            return;
        if (!(dialogButton.getParent() instanceof LinearLayout))
            return;            

        // Workaround for buttons with too long of text
        final LinearLayout linearLayout = (LinearLayout) dialogButton.getParent();
        linearLayout.addOnLayoutChangeListener(new View.OnLayoutChangeListener() {
            @Override
            public void onLayoutChange(View v, int left, int top, int right, int bottom, int oldLeft, int oldTop,
                                       int oldRight, int oldBottom) {
                if (right - left > 0) {
                    final int parentWidth = linearLayout.getWidth();
                    int childrenWidth = 0;
                    for (int i = 0; i < linearLayout.getChildCount(); ++i)
                        childrenWidth += linearLayout.getChildAt(i).getWidth();

                    if (childrenWidth > parentWidth) {
                        // Apply stacked buttons
                        linearLayout.setOrientation(LinearLayout.VERTICAL);
                        linearLayout.setPadding(linearLayout.getPaddingLeft(), 0, linearLayout.getPaddingRight(),
                                linearLayout.getPaddingBottom());
                        for (int i = 0; i < linearLayout.getChildCount(); ++i) {
                            if (linearLayout.getChildAt(i) instanceof Button) {
                                final Button child = (Button) linearLayout.getChildAt(i);
                                child.setGravity(Gravity.END | Gravity.CENTER_VERTICAL);
                                final LinearLayout.LayoutParams params = (LinearLayout.LayoutParams) child.getLayoutParams();
                                params.width = LinearLayout.LayoutParams.MATCH_PARENT;
                                params.gravity = Gravity.END;
                                child.setLayoutParams(params);
                            } else if (linearLayout.getChildAt(i) instanceof Space) {
                                linearLayout.removeViewAt(i--);
                            }
                        }
                    }

                    linearLayout.removeOnLayoutChangeListener(this);
                }
            }
        });
    }

    public static void requestPermission(int requestedPermissionInt) {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Requesting a permission...");

        String requestedPermission[] = getPermissionString(requestedPermissionInt);

        if (requestedPermission == null) {
            if (permissionCompleteCallback != null) {
                try {
                    permissionCompleteCallback.call();
                } catch (Exception e) {
                    Log.e(NoodleXBridge.TAG, "NATIVE~ Permission complete callback failed: " + e.toString());
                }
            }
            return;
        }

        if (havePermission(requestedPermissionInt)) {
            Log.d(NoodleXBridge.TAG, "NATIVE~ Permission has already been granted!");
            if (permissionCompleteCallback != null) {
                try {
                    permissionCompleteCallback.call();
                } catch (Exception e) {
                    Log.e(NoodleXBridge.TAG, "NATIVE~ Permission complete callback failed: " + e.toString());
                }
            }
            return;
        }

        pendingPermission = requestedPermissionInt;
        ActivityCompat.requestPermissions(mActivity.get(), requestedPermission, requestedPermissionInt);
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Permission successfully requested!");
    }

    public static void onRequestPermissionsResult(int requestCode, String permissions[], int[] grantResults) {
        if (permissionCompleteCallback != null) {
            try {
                permissionCompleteCallback.call();
            } catch (Exception e) {
                Log.e(NoodleXBridge.TAG, "NATIVE~ Permission complete callback failed: " + e.toString());
            }
        }
        return;
    }

    public static boolean havePermission(int requestedPermissionInt) {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Checking if we have a permission...");

        String requestedPermission[] = getPermissionString(requestedPermissionInt);

        if (requestedPermission == null) {
            return false;
        }

        if (Build.VERSION.SDK_INT < 23) {
            return true;
        }

        int permissionCheck = ContextCompat.checkSelfPermission(mActivity.get(), requestedPermission[0]);

        if (permissionCheck == mActivity.get().getPackageManager().PERMISSION_GRANTED) {
            return true;
        } else {
            return false;
        }
    }

    public static void setPermissionCompleteCallback(Callable<Void> callback) {
        permissionCompleteCallback = callback;
    }

    private static String[] getPermissionString(int requestedPermissionInt) {
        String requestedPermission[];

        switch (requestedPermissionInt) {
            case NoodleXBridge.NOODLEX_WRITE_EXTERNAL_STORAGE:
                requestedPermission = new String[]{Manifest.permission.WRITE_EXTERNAL_STORAGE};
                break;
            default:
                if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NATIVE~ Requested permission is not implemented!");
                return null;
        }
        return requestedPermission;
    }

    public static void setImmersiveMode() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.KITKAT) {
            //Set immersive mode
            mActivity.get().getWindow().getDecorView().setSystemUiVisibility(
                  View.SYSTEM_UI_FLAG_LAYOUT_STABLE
                | View.SYSTEM_UI_FLAG_LAYOUT_HIDE_NAVIGATION
                | View.SYSTEM_UI_FLAG_LAYOUT_FULLSCREEN
                | View.SYSTEM_UI_FLAG_HIDE_NAVIGATION
                | View.SYSTEM_UI_FLAG_FULLSCREEN
                | View.SYSTEM_UI_FLAG_IMMERSIVE_STICKY
                | View.SYSTEM_UI_FLAG_LOW_PROFILE);
        } else {
            mActivity.get().getWindow().getDecorView().setSystemUiVisibility(
                  View.SYSTEM_UI_FLAG_LAYOUT_STABLE
                | View.SYSTEM_UI_FLAG_LAYOUT_FULLSCREEN
                | View.SYSTEM_UI_FLAG_FULLSCREEN
                | View.SYSTEM_UI_FLAG_LOW_PROFILE);
        }
    }

    public static void keepScreenOn() {
        mActivity.get().getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);
    }

    public static String getAdvertisingID() {
        return advertisingID;
    }
}
