/****************************************************************************
 MADE BY LEALAND. BEN IS COOL. 27th, March, 2017

 This class handles NoodleNews.

  ### IMPORTANT SHOW CREATIVES INFO ###
  In order for NoodleNews to mute game audio for certain Creatives, a Callable has to be set up and linked between NoodleX and the game.
  How it Works:
    Basically you have to make a link (callable function) between NoodleX and the Game, so that when the willShow/didDismiss noodlenews delegate
    callback occur, they can then call muting/unmuting for the game.
    The NoodleNews callbacks flag local NoodleX variables for muting/unmuting, and the callback checks what those variables are when called game-side.
  
  To do this:
    1. Set the NoodleXNews callable, game-side, somewhere early in the games initialization stuff:
        // Set GPGS Cloud Load callback
        NoodleXBridge.setCreativesCallable(new Callable<Void>()
            {
                public Void call()
                {   
                    noodlenewsMuteAudio();    // <--- this is the game-side function that will mute/unmute the game, implemented via 2. bellow
                    return null;
                }
            });

    2. Implement the game-side function called above:
            // This method handles muting/unmuting for when noodlenews creatives are played
            private static boolean mutedForCreative = false;  //  True when the audio was muted for a creative. Unmute on creative dismissed. Prevent unmuting when muted before creative
            public static void noodlenewsMuteAudio() {
                // 
                if (NoodleXBridge.noodlenewsMuteAudio() && !gameMuted)
                {
                    Log.d(Extension.EXT_LOG_TAG, "NEWS~ Game muted for creative!");
                    // Mute game audio
                    // MUTE GAME
                    // Flag that game was muted for creative
                    mutedForCreative = true;
                }
                else
                {
                    Log.d(Extension.EXT_LOG_TAG, "NEWS~ Check to unmute game.");
                    // Check to unmute
                    if (mutedForCreative)
                    {
                        Log.d(Extension.EXT_LOG_TAG, "NEWS~ Game unmuted after creative!");
                        // Unmute game Audio
                        // UNMUTE GAME
                        // Flag no longer muted for creative
                        mutedForCreative = false;
                    }
                }
            }

 ****************************************************************************/

package com.noodlecake.NoodleX;

import java.util.*;
import java.util.concurrent.Callable;

import java.io.IOException;

import java.lang.Object;
import java.lang.ref.WeakReference;

import java.security.spec.EllipticCurve;
import javax.security.auth.callback.Callback;

import android.util.Log;

import android.app.Activity;
import android.app.Application;

import android.os.Bundle;
import android.os.AsyncTask;

import android.content.Intent;
import android.content.IntentSender;
import android.content.Context;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;

import com.google.android.gms.games.Games;
import com.google.android.gms.common.*;

import com.noodlecake.noodlenews.promotion.Creative;
import com.noodlecake.noodlenews.NoodleNewsClient;

import com.noodlecake.NoodleX.NoodleXBridge;

public class NoodleXNews
    implements com.noodlecake.noodlenews.INoodleNewsClientDelegate
{
    /*
     * VARIABLES
     */
    // Core stuff
    private static WeakReference<Activity> mActivity;                   // A weak reference to this app's Activity
    private static NoodleXNews mInstance;                               // The instance of this class, needed for recieveing callbacks
    // Preferences
    private static SharedPreferences preferences;                       // The Preferences for this app. Stores relevant settings
    private static SharedPreferences.Editor prefEditor;                 // Used to edit the shared preferences. 
    // Parameters
    private static int minLaunches = 5;                                 // The minimum number of game launches before an exposure will be shown (greater than).
    // Audio control 
    private static Callable<Void> creativeAudioCallable;                // Game-side method to mute/unmute for Noodlenews creatives
    private static boolean muteAudio;                                   // True when the in-game audio should be muted.

    /*********************
       NOODLE NEWS STUFF
     *********************/
    public static void init (WeakReference<Activity> activity)
    {
        Log.d(NoodleXBridge.TAG, "NEWS~ Initializing NoodleXNews...");
        if (mInstance != null)
        {
            Log.d(NoodleXBridge.TAG, "NEWS~ NoodleXNews already initialized.");
            return;
        }
        
        // Create an instance
        mInstance = new NoodleXNews();
        
        // Load noodlenews library
        System.loadLibrary("nativeNoodleNews");
        
        // Set up Preferences
        preferences = activity.get().getPreferences(Context.MODE_PRIVATE);
        // Set up preference editor
        prefEditor = preferences.edit();

        // Connect noodlenews 
        NoodleXNews.startNoodleNewsSession(activity.get());
        // Setup callbacks
        NoodleNewsClient.setDelegate(mInstance);

        Log.d(NoodleXBridge.TAG, "NEWS~ NoodleXNews initialization complete!");
    }

    
    // This method attemps to start a noodlenews session
    public static void startNoodleNewsSession (Activity activity)
    {
        NoodleNewsClient.startSession(activity, "google");
    }


    /*************
       CREATIVES
     *************/
    // This method is called by game-side reativeAudioController
    public static boolean noodlenewsMuteAudio ()
    {
        return muteAudio;
    }


    // This method sets the callback for controlling audio during creatives
    public static void setCreativesCallable (Callable<Void> callable)
    {   
        creativeAudioCallable = callable;
    }


    // This method returns whether or not an exposure can be shown.
    public static boolean canShowCreative ()
    {
        if (preferences == null)
        {
            Log.d(NoodleXBridge.TAG, "NEWS~ Cannot show exposure, preferences not initialized..");
            return false;
        }

        //Check how many game launches have occured
        boolean showCreative = false;
        // Check if exposure is ready
        if (NoodleNewsClient.hasPendingCreative()) showCreative = true;

        // Check game-side parameters
        if (NoodleXBridge.getLaunchCount() < minLaunches) showCreative = false;

        return showCreative;
    }


    // callback for when a creative is shown.
    public void willShowCreative(Creative creative, boolean shouldMuteAudio)
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NEWS~ showing creative, shouldMute: "+shouldMuteAudio);
        // flag audio for mute
        muteAudio = shouldMuteAudio;
        // Call game-side audion controller
        try
        {
            if (creativeAudioCallable != null) creativeAudioCallable.call();
        }
        catch (Exception e)
        {
            Log.d(NoodleXBridge.TAG, "NEWS~ Exception in dismiss creative callable: "+e.toString());
        }
    }


    // callback for when a creative is dismissed.
    public void didDismissCreative(Creative creative, boolean tookPositiveAction)
    {
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NEWS~ dismissing creative: ");
        // Flag immersive mode
        NoodleXBridge.setImmersiveMode();
        // Flag audio for unmute
        muteAudio = false;
        // Call game-side audion controller
        try
        {
            if (creativeAudioCallable != null) creativeAudioCallable.call();
        }
        catch (Exception e)
        {
            Log.d(NoodleXBridge.TAG, "NEWS~ Exception in dismiss creative callable: "+e.toString());
        }
    }


    // callback for when a creative is dismissed.
    public static void showNoodleNewsCreative ()
    {
        if (canShowCreative()) NoodleNewsClient.showCreative();
    }


    /**************
       More Games
     **************/
    // Callback method for when More Games will be shown.
    public void willShowMoreGames ()
    {
        Log.d(NoodleXBridge.TAG, "NEWS~ noodlenews ID: "+NoodleNewsClient.getSupportIdentifier());
        if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NEWS~ WillShowMoreGames delegate called.");
    }

    // Callback method for when more games was dismissed.
    public void didDismissMoreGames ()
    {
         if (NoodleXBridge.debug) Log.d(NoodleXBridge.TAG, "NEWS~ Dismissing More Games, and setting immersive mode");
        NoodleXBridge.setImmersiveMode();
    }

    // Shows more games via NoodleNews.
    public static void showMoreGames ()
    {
        NoodleNewsClient.showMoreGames();
    }

}   // End of class